# -*- coding: utf-8 -*-
"""Common routines for reading maximus files."""
__all__ = ["read_scan", "hdr_to_dict", "flatten_header", "process_key", "read_images", "read_pointscan"]
import json
from pathlib import Path
import re

import numpy as np

from ...tools.file import FileManager


def read_scan(file_root):
    """Find the .hdr and .xim/,xsp files for a scan load them into memory.

    Args:
        file_root (str): This is the part of the hdr filename beore the extension.

    Returns:
        (dict,ndarray, (n-1d arrays)):
            Returns the metadata and an ndarray of the scan data and then n 1D arrays of the axes.
    """
    hdr = Path(str(file_root) + ".hdr")
    header = hdr_to_dict(hdr, to_python=True)
    scan_type = header["ScanDefinition"]["Type"]

    if "Image Scan" in scan_type:
        data, dims = read_images(hdr.parent.glob(f"{hdr.stem}*.xim"), header)
    elif "Point Scan" in scan_type:
        data, dims = read_pointscan(hdr.parent.glob(f"{hdr.stem}*.xsp"), header)
    else:
        raise ValueError(f"Unrecognised scan type {scan_type}")

    return header, data, dims


def hdr_to_dict(filename, to_python=True):
    """Convert .hdr metadata file to  json or python dictionary.

    Args:
        filename (str):
            Name of file to read (can also be a pathlib.Path).

    Keyword Arguments:
        to_python (bool):
            If true, return a python dictionary, otherwise return a json text string.

    Returns:
        (dict or str):
            Either the header file as a python dictionary, or a json string.
    """
    bare = re.compile(r"([\s\{])([A-Za-z][A-Za-z0-9_]*)\s\:")  # Match for keys
    term = re.compile(r",\s*([\]\}])")  # match for extra , at the end of a dict or list
    nan = re.compile(r"([\-0-9\.]+\#QNAN)")  # Handle NaN values

    # Use oathlib to suck in the file
    with FileManager(filename, "r") as f:
        hdr = f.read()
    # Simple string replacements first
    stage1 = hdr.replace("=", ":").replace(";", ",").replace("(", "[").replace(")", "]")
    # Wrap in { }
    stage2 = f"{{{stage1}}}"
    # Regexp replacements next
    stage3 = bare.sub('\\1"\\2" :', stage2)
    stage4 = term.sub("\\n\\1", stage3)
    stage5 = nan.sub("NaN", stage4)

    if to_python:
        ret = process_key(json.loads(stage5))
    else:  # orettyify the json
        ret = json.dumps(json.loads(stage5), indent=4)

    return ret


def flatten_header(value):
    """Flatten nested dictionaries."""
    if isinstance(value, list) and len(value) == 1:
        value = value[0]
    if not isinstance(value, dict):
        return value
    dels = []
    adds = {}
    for key, item in value.items():
        item = flatten_header(item)
        if isinstance(item, dict):
            for item_key, item_val in item.items():
                adds[f"{key}.{item_key}"] = item_val
            dels.append(key)
    for key in dels:
        del value[key]
    value.update(adds)
    return value


def process_key(value):
    """Carry out post loading processing of data structures."""
    if isinstance(value, dict):
        for key, val in value.items():
            value[key] = process_key(val)
        return value
    if isinstance(value, list):
        if len(value) > 0 and isinstance(value[0], int) and len(value) == value[0] + 1:  # Lists prepended with size
            del value[0]
        a = np.array(value)
        if a.dtype.kind in ["f", "i", "u", "U"]:  # convert arrays to arrays
            return a
        value = [process_key(v) for v in value]
    return value


def read_images(files, header):
    """Read one or more .xim files and construct the data array.

    Args:
        files (glob): glob pattern of xim files to read.
        header (dict): contents of the .jdr file.

    Returns:
        data (ndarray): 2D or 3D data.
        dims (tuple of 1D arrays): 2 or 3 1D arrays corresponding to the dimensions of data.
    """
    xims = list(files)
    scandef = header["ScanDefinition"]
    region = scandef["Regions"][0]  # FIXME assumes a single region in the data
    if len(xims) > 1:
        data = np.stack([np.genfromtxt(x)[::-1] for x in xims]).T
    elif len(xims) == 1:
        data = np.genfromtxt(xims[0])[::-1]
    else:  # no files !
        raise IOError("No Images located")
    xpts = region["PAxis"]["Points"]
    ypts = region["QAxis"]["Points"]
    if data.ndim == 3:
        zpts = scandef["StackAxis"]["Points"]
        dims = (xpts, ypts, zpts)
    else:
        dims = (xpts, ypts)
    return data, dims


def read_pointscan(files, header):
    """Read one or more .xsp files and construct the data array.

    Args:
        files (glob): glob pattern of xim files to read.
        header (dict): contents of the .jdr file.

    Returns:
        data (ndarray): 2D or 3D data.
        dims (tuple of 1D arrays): 2 or 3 1D arrays corresponding to the dimensions of data.
    """
    xsps = list(files)
    scandef = header["ScanDefinition"]
    region = scandef["Regions"][0]  # FIXME assumes a single region in the data
    if len(xsps) > 1:
        data = np.stack([np.genfromtxt(x)[:, 1] for x in xsps]).T
    elif len(xsps) == 1:
        data = np.genfromtxt(xsps[0])[:, 1]
    else:  # No files !
        raise IOError("No Spectra located")
    xpts = region["PAxis"]["Points"]
    if data.ndim == 2:
        zpts = scandef["StackAxis"]["Points"]
        dims = (xpts, zpts)
    else:
        dims = (xpts,)
    return data, dims



# -*- coding: utf-8 -*-
# pylint: disable=no-member, invalid-name
"""Demonstrates the interactive mask selection code."""

from Stoner import __homepath__
from Stoner.HDF5 import STXMImage

# Load a single helicity STXM image
img = STXMImage(
    __homepath__ / ".." / "sample-data" / "Sample_Image_2017-10-15_100.hdf5"
)

# USe the interactive mask selection function
img.mask.select()

# Normalise the image based on the masked region
img.normalise(limits=(0.01, 0.99), scale_masked=True)
img.imshow()

# Quantize the image colours based on a threshold calculated from the image.
img.clone.quantize(output=[-1, 1], levels=[img.threshold_otsu()]).imshow()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Utility functions to support :py:mod:`Stoner.Core`."""

__all__ = ["add_core", "and_core", "sub_core", "mod_core", "copy_into", "tab_delimited", "decode_string"]

import copy
import csv
import re
from collections.abc import Mapping
from typing import Union, List, Mapping as MappingType, Callable
import numpy as np
from ..compat import index_types, int_types
from ..tools import all_type, copy_into
from .Typing import Numeric, Column_Index, Int_Types


def add_core(other: Union["DataFile", np.ndarray, List[Numeric], MappingType], newdata: "DataFile") -> "DataFile":
    """Implement the core work of adding other to self and modifying newdata.

    Args:
        other (DataFile,array,list):
            The data to be added
        newdata(DataFile):
            The instance to be modified

    Returns:
        newdata:
            A modified newdata
    """
    if isinstance(other, np.ndarray):
        if len(newdata) == 0:  # pylint: disable=len-as-condition
            ch = getattr(other, "column_headers", [])
            setas = getattr(other, "setas", "")
            t = np.atleast_2d(other)
            c = t.shape[1]
            if len(newdata.column_headers) < c:
                newdata.column_headers.extend([f"Column_{x}" for x in range(c - len(newdata.column_headers))])
            newdata.data = t
            newdata.setas = setas
            newdata.column_headers = ch
            ret = newdata
        elif len(np.shape(other)) == 1:
            # 1D array, so assume a single row of data
            if np.shape(other)[0] == np.shape(newdata.data)[1]:
                newdata.data = np.append(newdata.data, np.atleast_2d(other), 0)
                ret = newdata
            else:
                return NotImplemented
        elif len(np.shape(other)) == 2 and np.shape(other)[1] == np.shape(newdata.data)[1]:
            # DataFile + array with correct number of columns
            newdata.data = np.append(newdata.data, other, 0)
            ret = newdata
        else:
            return NotImplemented
    elif isinstance(other, type(newdata)):  # Appending another DataFile
        new_data = np.ones((other.shape[0], newdata.shape[1])) * np.nan
        for i in range(newdata.shape[1]):
            column = newdata.column_headers[i]
            try:
                new_data[:, i] = other.column(column)
            except KeyError:
                pass
        newdata.metadata.update(other.metadata)
        newdata.data = np.append(newdata.data, new_data, axis=0)
        ret = newdata
    elif isinstance(other, list):
        for o in other:
            newdata = newdata + o
        ret = newdata
    elif isinstance(other, Mapping):
        # First check keys all in newdata
        if len(newdata) == 0:
            newdata.data = np.atleast_2d(list(other.values()))
            newdata.column_headers = list(other.keys())
        else:
            order = dict()
            for k in other:
                try:
                    order[k] = newdata.find_col(k)
                except (KeyError, re.error):
                    mask = newdata.mask
                    newdata.add_column(np.ones(len(newdata)) * np.NaN, header=k)
                    newdata.mask[:, :-1] = mask
                    newdata.mask[:, -1] = np.ones(len(newdata), dtype=bool)
                    order[k] = newdata.shape[1] - 1
            row = np.ones(newdata.shape[1]) * np.NaN
            mask = np.ones_like(row, dtype=bool)
            for k in order:
                row[order[k]] = other[k]
                mask[order[k]] = False
            old_mask = newdata.mask
            newdata.data = np.ma.append(newdata.data, np.atleast_2d(row), axis=0)
            newdata.mask[:-1, :] = old_mask
            newdata.mask[-1] = mask
        ret = newdata
    else:
        return NotImplemented
    ret._data._setas.shape = ret.shape
    for attr in newdata.__dict__:
        if attr not in ("setas", "metadata", "data", "column_headers", "mask") and not attr.startswith("_"):
            ret.__dict__[attr] = newdata.__dict__[attr]
    return ret


def and_core(other: Union["DataFile", np.ndarray], newdata: "DataFile") -> "DataFile":
    """Implement the core of the & operator, returning data in newdata.

    Args:
        other (array,DataFile):
            Data whose columns are to be added
        newdata (DataFile):
            instance of DataFile to be modified

    Returns:
        ():py:class:`DataFile`):
            new Data object with the columns of other concatenated as new columns at the end of the self object.
    """
    if len(newdata.data.shape) < 2:
        newdata.data = np.atleast_2d(newdata.data)

    # Get other to be a numpy masked array of data
    # Get other_headers to be a suitable length list of strings
    if isinstance(other, type(newdata)):
        newdata.metadata.update(other.metadata)
        other_headers = other.column_headers
        other = copy.copy(other.data)
    elif isinstance(other, type(newdata.data)):
        other = copy.copy(other)
        if other.ndim < 2:  # 1D array, make it 2D column
            other = np.atleast_2d(other)
            other = other.T
        other_headers = [f"Column {i + newdata.shape[1]}" for i in range(other.shape[1])]
    elif isinstance(other, np.ndarray):
        other = type(newdata.data)(copy.copy(other))
        if other.ndim < 2:  # 1D array, make it 2D column
            other = np.atleast_2d(other)
            other = other.T
        other_headers = [f"Column {i + newdata.shape[1]}" for i in range(other.shape[1])]
    else:
        return NotImplemented

    newdata_headers = newdata.column_headers + other_headers
    setas = newdata.setas.clone

    # Workout whether to extend rows on one side or the other
    if np.prod(newdata.data.shape) == 0:  # Special case no data yet
        newdata.data = other
    elif newdata.data.shape[0] == other.shape[0]:
        newdata.data = np.append(newdata.data, other, 1)
    elif newdata.data.shape[0] < other.shape[0]:  # Need to extend self.data
        extra_rows = other.shape[0] - newdata.data.shape[0]
        newdata.data = np.append(newdata.data, np.zeros((extra_rows, newdata.data.shape[1])), 0)
        new_mask = newdata.mask
        new_mask[-extra_rows:, :] = True
        newdata.data = np.append(newdata.data, other, 1)
        other_mask = np.ma.getmaskarray(other)
        new_mask = np.append(new_mask, other_mask, 1)
        newdata.mask = new_mask
    elif other.shape[0] < newdata.data.shape[0]:
        # too few rows we can extend with zeros
        extra_rows = newdata.data.shape[0] - other.shape[0]
        other = np.append(other, np.zeros((extra_rows, other.shape[1])), 0)
        other_mask = np.ma.getmaskarray(other)
        other_mask[-extra_rows:, :] = True
        new_mask = newdata.mask
        new_mask = np.append(new_mask, other_mask, 1)
        newdata.data = np.append(newdata.data, other, 1)
        newdata.mask = new_mask

    setas.column_headers = newdata_headers
    newdata._data._setas = setas
    newdata._data._setas.shape = newdata.shape
    for attr in newdata.__dict__:
        if attr not in ("setas", "metadata", "data", "column_headers", "mask") and not attr.startswith("_"):
            newdata.__dict__[attr] = newdata.__dict__[attr]
    return newdata


def mod_core(other: Column_Index, newdata: "DataFile") -> "DataFile":
    """Implement the column deletion method."""
    if isinstance(other, index_types):
        newdata.del_column(other)
    else:
        newdata = NotImplemented
    newdata._data._setas.shape = newdata.shape
    return newdata


def sub_core(other: Union[Int_Types, slice, Callable], newdata: "DataFile") -> "DataFile":
    """Worker for the subtraction."""
    if isinstance(other, (slice, int_types)) or callable(other):
        newdata.del_rows(other)
    elif isinstance(other, list) and (all_type(other, int_types) or all_type(other, bool)):
        newdata.del_rows(other)
    else:
        newdata = NotImplemented
    newdata._data._setas.shape = newdata.shape
    return newdata


class tab_delimited(csv.Dialect):
    """A customised csv dialect class for reading tab delimited text files."""

    delimiter = "\t"
    quoting = csv.QUOTE_NONE
    doublequote = False
    lineterminator = "\r\n"


def decode_string(value: str) -> str:
    """Expand a string of column assignments, replacing numbers with repeated characters."""
    pattern = re.compile(r"(([0-9]+)(x|y|z|d|e|f|u|v|w|\.|\-))")
    while True:
        res = pattern.search(value)
        if res is None:
            break
        (total, count, code) = res.groups()
        count = int(count)
        value = value.replace(total, code * count, 1)
    return value



# -*- coding: utf-8 -*-
"""Provide variants of :class:`Stoner.Image.ImageFolder` that store images efficiently in 3D numpy arrays."""
__all__ = ["ImageStackMixin", "ImageStack", "ImageStack"]
import warnings

import numpy as np

from ..compat import string_types, int_types
from ..core.exceptions import assertion

from ..Core import regexpDict, typeHintedDict
from ..Folders import DiskBasedFolderMixin, baseFolder

from .core import ImageArray, ImageFile
from .folders import ImageFolder, ImageFolderMixin

IM_SIZE = (512, 672)  # Standard Kerr image size
AN_IM_SIZE = (554, 672)  # Kerr image with annotation not cropped


def _load_ImageArray(f, **kargs):
    """Create and image array."""
    kargs.pop("Img_num", None)  # REemove img_num if it exists
    return ImageArray(f, **kargs)


class ImageStackMixin:
    """Implement an interface for a baseFolder to store images in a 3D numpy array for faster access."""

    _defaults = {"type": ImageFile}

    def __init__(self, *args, **kargs):
        """Initialise an ImageStack's pricate data and provide a type argument."""
        self._stack = np.ma.atleast_3d(ImageArray([])).reshape((0, 0, 0)).view(ImageArray)
        self._metadata = regexpDict()
        self._public_attrs_store = regexpDict()
        self._names = list()
        self._sizes = np.array([], dtype=int).reshape(0, 2)

        if not len(args):
            super().__init__(**kargs)
            return None  # No further initialisation
        other = args[0]
        if isinstance(other, ImageStackMixin):
            super().__init__(*args[1:], **kargs)
            self._stack = other._stack
            self._metadata = other._metadata
            self._names = other._names
            self._sizes = other._sizes
        elif isinstance(other, ImageFolder):  # ImageFolder can already init from itself
            super().__init__(*args, **kargs)
        elif (
            isinstance(other, np.ndarray) and other.ndim == 3
        ):  # Initialise with 3D numpy array, first coordinate is number of images
            super().__init__(*args[1:], **kargs)
            self.imarray = other
            self._sizes = np.ones((other.shape[0], 2), dtype=int) * other.shape[1:]
            self._names = [f"Untitled-{d}" for d in range(other.shape[0])]
            for n in self._names:
                self._metadata[n] = typeHintedDict()
        elif isinstance(other, list):
            try:
                other = [ImageFile(i) for i in other]
            except (TypeError, ValueError, RuntimeError) as err:
                raise ValueError("Failed to initialise ImageStack with list input") from err
            super().__init__(*args[1:], **kargs)
            for ot in other:
                self.append(ot)

        else:
            super().__init__(*args, **kargs)

    def __lookup__(self, name):
        """Stub for other classes to implement.

        Parameters:
            name(str): Name of an object

        Returns:
            A key in whatever form the :py:meth:`baseFolder.__getter__` will accept.

        Note:
            We're in the base class here, so we don't call super() if we can't handle this, then we're stuffed!
        """
        if isinstance(name, int_types):
            try:
                self._stack[:, :, name]
            except IndexError as err:
                raise KeyError(f"{name} is out of range for accessing the ImageStack.") from err
            return name
        if name not in self.__names__():
            name = self._metadata.__lookup__(name)
        return list(self._metadata.keys()).index(name)  # return the matching index of the name

    def __names__(self):
        """Stub method to return a list of names of all objects that can be indexed for __getter__.

        Note:
            We're in the base class here, so we don't call super() if we can't handle this, then we're stuffed!
        """
        return list(self._metadata.keys())

    def __getter__(self, name, instantiate=True):
        """Stub method to do whatever is needed to transform a key to a metadataObject.

        Parameters:
            name (key type): The canonical mapping key to get the dataObject. By default
                the baseFolder class uses a :py:class:`regexpDict` to store objects in.

        Keyword Arguments:
            instantiate (bool): If True (default) then always return a metadataObject. If False,
                the __getter__ method may return a key that can be used by it later to actually get the
                metadataObject. If None, then will return whatever is helf in the object cache, either
                instance or name.

        Returns:
            (metadataObject): The metadataObject

            Note:
            We're in the base class here, so we don't call super() if we can't handle this, then we're stuffed!


        """
        try:
            idx = self.__lookup__(name)
        except KeyError:
            # If we don't seem to have the name then see if we can fall back to something else like a
            #  DiskBasedFolderMixin
            return super().__getter__(name, instantiate)
        if isinstance(instantiate, bool) and not instantiate:
            return self.__names__()[idx]
        instance = self._instantiate(idx)
        return self._update_from_object_attrs(instance)

    def __setter__(self, name, value, force_insert=False):
        """Stub to setting routine to store a metadataObject.

        Parameters:
            name (string):
                the named object to write - may be an existing or new name
            value (metadataObject):
                the value to store.

        Note:
            We're in the base class here, so we don't call super() if we can't handle this, then we're stuffed!
        """
        if isinstance(name, int_types):
            try:
                name = self.__names__()[name]
            except IndexError:
                name = self.make_name(value)
        if name is None:
            name = self.make_name(value)
        try:
            if force_insert:
                raise KeyError("Fake force insert")
            idx = self.__lookup__(name)
        except KeyError:  # Ok we're appending here
            if isinstance(value, string_types):  # Append with a filename, call __getter__
                value = self.__getter__(value, instantiate=True)  # self.__getter__ will also insert if necessary
                return None
            idx = len(self)
            return self.__inserter__(idx, name, value)
        if not isinstance(value, self.type):
            value = self.type(value)
        self._sizes[idx] = value.shape
        self._metadata[name] = value.metadata
        _public_attrs = {}
        for attr in value._public_attrs.keys():
            if attr in ["data", "image", "metadata"]:
                continue  # skip attrs I already handle
            _public_attrs[attr] = getattr(value, attr, None)
        self._public_attrs_store[name] = _public_attrs
        value = getattr(value, "image", value)
        mask = getattr(value, "mask", np.zeros_like(value, dtype=bool))
        row, col = value.shape
        pag = len(self._sizes)
        new_size = self.max_size + (pag,)
        if new_size[2] == 1:
            dtype = value.dtype
        else:
            dtype = None
        self._resize_stack(new_size, dtype=dtype)
        self._stack[:row, :col, idx] = value
        self._stack.mask = np.ma.getmaskarray(self._stack)
        self._stack.mask[:row, :col, idx] = mask

    def __inserter__(self, ix, name, value):
        """Provide an efficient insert into the stack.

        The default implementation is rather slow about inserting since it has to clear the data folder and then
        rebuild it entry by entry. This does
        a simple insert."""
        value = ImageFile(value)  # ensure we have some metadata
        self._names.insert(ix, name)
        self._metadata[name] = value.metadata
        self._sizes = np.insert(self._sizes, ix, value.shape, axis=0)  # pylint: disable=no-member
        new_size = self.max_size + (len(self._names),)
        if new_size[2] == 1:
            dtype = value.image.dtype
        else:
            dtype = None
        self._resize_stack(new_size, dtype=dtype)
        stack_mask = np.insert(np.ma.getmaskarray(self._stack), ix, np.zeros(self.max_size, dtype=bool), axis=2)
        self._stack = np.insert(self._stack, ix, np.zeros(self.max_size), axis=2).view(ImageArray)
        self._stack.mask = stack_mask
        self._stack = self._stack[: new_size[0], : new_size[1], : new_size[2]]
        value = value.data
        row, col = value.shape
        self._stack[:row, :col, ix] = value
        self._stack.mask[:row, :col, ix] = np.ma.getmaskarray(value)
        _public_attrs = {}
        for attr in value._public_attrs.keys():
            if attr in ["data", "image", "metadata"]:
                continue  # skip attrs I already handle
            _public_attrs[attr] = getattr(value, attr, None)
        self._public_attrs_store[name] = _public_attrs

    def __deleter__(self, ix):
        """Delete an object from the baseFolder.

        Parameters:
            ix(str): Index to delete, should be within +- the lengthe length of the folder.

        Note:
            We're in the base class here, so we don't call super() if we can't handle this, then we're stuffed!

        """
        idx = self.__lookup__(ix)
        name = list(self.__names__())[idx]
        del self._metadata[name]
        self._stack = np.delete(self._stack, idx, axis=2)
        del self._names[idx]
        self._sizes = np.delete(self._sizes, ix, axis=0)

    def __clear__(self):
        """Clear all stored :py:class:`Stoner.Core.metadataObject` instances stored.

        Note:
            We're in the base class here, so we don't call super() if we can't handle this, then we're stuffed!

        """
        self._metadata = regexpDict()
        self._stack = np.ma.atleast_3d(ImageArray([])).reshape((0, 0, 0)).view(ImageArray)

    ###########################################################################
    ###################      Special methods     ##############################

    def __floordiv__(self, other):
        """Calculate and XMCD ratio on the images."""
        if not isinstance(other, ImageStackMixin):
            return NotImplemented
        if self._stack.dtype != other._stack.dtype:
            raise ValueError(
                "Only ImageFiles with the same type of underlying image data can be used to calculate an XMCD ratio."
                + "Mismatch is {self._stack.dtype} vs {other._stack.dtype}"
            )
        if self._stack.dtype.kind != "f":
            ret = self.clone.convert(float)
            other = other.clone.convert(float)
        else:
            ret = self.clone
        if other._stack.shape != ret._stack.shape:
            raise ValueError(
                "Only image stacks that are the same shape can be used to calaculate XMCD data"
                + f" - passed {ret._stack.shape} and {other._stack.shape}"
            )
        ret._stack = (ret._stack - other._stack) / (ret._stack + other._stack)

        return ret

    ###########################################################################
    ###################      Private methods     ##############################

    def _instantiate(self, idx):
        """Reconstructs the data type."""
        r, c = self._sizes[idx]
        if issubclass(
            self.type, ImageArray
        ):  # IF the underlying type is an ImageArray, then return as a view with extra metadata
            tmp = self._stack[:r, :c, idx].view(type=self.type)
        else:  # Otherwise it must be something with a data attribute
            tmp = self.type(self._stack[:r, :c, idx])
        name = self.__names__()[idx]
        tmp.metadata = self._metadata[name]
        tmp._fromstack = True
        if name in self._public_attrs_store:
            for attr, value in self._public_attrs_store[name].items():
                setattr(tmp, attr, value)
        if tmp.filename == "":
            tmp.filename = self.__names__()[idx]
        return tmp

    def _resize_stack(self, new_size, dtype=None):
        """Create a new stack with a new size."""
        old_size = self._stack.shape
        assertion(isinstance(self._stack, ImageArray), f"Trying to resize a non-image array {type(self._stack)}")
        if old_size == new_size:
            return new_size
        if dtype is None:
            dtype = self._stack.dtype
        row, col, pag = tuple([min(o, n) for o, n in zip(old_size, new_size)])

        new = np.ma.zeros(new_size, dtype=dtype).view(ImageArray)
        new[:row, :col, :pag] = self._stack[:row, :col, :pag]
        new.mask = np.ma.getmaskarray(self._stack)[:row, :col, :pag]
        self._stack = new
        return row, col, pag

    ###########################################################################
    ################### Properties of ImageStack ##############################

    @property
    def imarray(self):
        """Produce the 3D stack of images - as [image,x,y]."""
        return np.transpose(self._stack, (2, 0, 1))

    @imarray.setter
    def imarray(self, value):
        """Set the 3D stack of images - as [image,x,y]."""
        value = np.ma.atleast_3d(value)
        self._stack = np.transpose(value, (1, 2, 0)).view(ImageArray)

    @property
    def max_size(self):
        """Get the biggest image dimensions in the stack."""
        if np.prod(self._sizes.shape) == 0:
            return (0, 0)
        return (self._sizes[:, 0].max(), self._sizes[:, 1].max())

    @property
    def shape(self):
        """Return the stack shape - after re-ordering the indices."""
        x, y, z = self._stack.shape
        return (z, x, y)

    ###########################################################################
    ###################         Public  methods         #######################

    def convert(self, dtype, force_copy=False, uniform=False, normalise=True):
        """Convert an image to the requested data-type.

        Warnings are issued in case of precision loss, or when negative values
        are clipped during conversion to unsigned integer types (sign loss).

        Floating point values are expected to be normalized and will be clipped
        to the range [0.0, 1.0] or [-1.0, 1.0] when converting to unsigned or
        signed integers respectively.

        Numbers are not shifted to the negative side when converting from
        unsigned to signed integer types. Negative values will be clipped when
        converting to unsigned integers.

        Parameters:
            image (ndarray):
                    Input image.
            dtype (dtype)
                Target data-type.
            force_copy (bool):
                Force a copy of the data, irrespective of its current dtype.
            uniform (bool):
                Uniformly quantize the floating point range to the integer range.
                By default (uniform=False) floating point values are scaled and
                rounded to the nearest integers, which minimizes back and forth
                conversion errors.
            normalise (bool):
                When converting from int types to float normalise the resulting array
                by the maximum allowed value of the int type.

        References:
            1.  DirectX data conversion rules.
                http://msdn.microsoft.com/en-us/library/windows/desktop/dd607323%28v=vs.85%29.aspx
            2,  Data Conversions.
                In "OpenGL ES 2.0 Specification v2.0.25", pp 7-8. Khronos Group, 2010.
            3,  Proper treatment of pixels as integers. A.W. Path.
                In "Graphics Gems I", pp 249-256. Morgan Kaufmann, 1990.
            4,  Dirty Pixels. J. Blinn.
                In "Jim Blinn's corner: Dirty Pixels", pp 47-57. Morgan Kaufmann, 1998.
        """
        from .imagefuncs import convert

        # Actually this is just a pass through for the imagefuncs.convert routine
        mask = self._stack.mask
        self._stack = convert(self._stack, dtype, force_copy=force_copy, uniform=uniform, normalise=normalise).view(
            type(self._stack)
        )
        self._stack.mask = mask
        return self

    def asfloat(self, normalise=True, clip=False, clip_negative=False, **kargs):
        """Convert stack to floating point type.

        Keyword Arguments:
            normalise(bool):
                normalise the image to the max value of current int type
            clip(bool):
                clip resulting range to values between -1 and 1
            clip_negative(bool):
                clip range further to 0,1

        Notes:
            Analogous behaviour to ImageFile.asfloat()

            If currently an int type and normalise then floats will be normalised
            to the maximum allowed value of the int type.
            If currently a float type then no change occurs.
            If clip_negative then clip values outside the range 0,1

        """
        if self.imarray.dtype.kind == "f":
            pass
        else:
            self.convert(dtype=np.float64, normalise=normalise)
        if "clip_neg" in kargs:
            warnings.warn(
                "clip_neg argument renamed to clip_negative in ImageStack. This will cause an error in future"
                + "versions of the Stoner Package."
            )
            clip_negative = kargs.pop("clip_neg")
        if clip or clip_negative:
            self.each.clip_intensity(clip_negative=clip_negative)
        return self

    def dtype_limits(self, clip_negative=True):
        """Return intensity limits, i.e. (min, max) tuple, of imarray dtype.

        Keyword Arguments:
            clip_negative(bool):
                If True, clip the negative range (i.e. return 0 for min intensity)
                even if the image dtype allows negative values.

        Returns:
            (imin,imax) (tuple):
                Lower and upper intensity limits.
        """
        ret = self[0].dtype_limits
        if clip_negative:
            ret = [max(0, x) for x in ret]
        return ret

    ###########################################################################
    ################### Deprecated Compatibility methods #######################

    def correct_drifts(self, refindex, threshold=0.005, upsample_factor=50, box=None):
        """Align images to correct for image drift.

        Pass through to ImageArray.corret_drift.

        Arg:
            refindex: int or str
                index or name of the reference image to use for zero drift
        Keyword Arguments:
            threshold(float): see ImageArray.correct_drift
            upsample_factor(int): see ImageArray.correct_drift
            box: see ImageArray.correct_drift

        """
        warnings.warn("correct_drift is a deprecated method for an image stack - consider using align.")
        ref = self[refindex]
        self.apply_all("correct_drift", ref, threshold=threshold, upsample_factor=upsample_factor, box=box)

    def crop_stack(self, box):
        """Crop the imagestack to a box.

        Args:
            box(array or list of type int):
                [xmin,xmax,ymin,ymax]

        Returns:
            (ImageStack):
                cropped images
        """
        warnings.warn("crop_stack is deprecated - sam effect can be achieved with crop(box)")
        self.each.crop(box)

    def show(self):
        """Pass through to :py:meth:`Stoner.Image.ImageFolder.view`."""
        warnings.warn("show() is deprecated in favour of ImageFolder.view()")
        return self.view()


class StackAnalysisMixin:
    """Add some analysis capability to ImageStack.

    These functions may override :py:class:`Stoner,Image.ImageFile` functions but do them efficiently for a numpy
    stack of images.
    """

    def subtract(self, background):
        """Subtract a background image (or index) from all images in the stack.

        Arguments:
            background (int, str or 2D array):
                Background image to index

        Returns:
            (ImageStack):
                The modified image stack.

        Notes:
            Method changed for v0.10 to not normalise or clip the data.
            The background image is scaled by the ratio of the mean pixel values of the unmasked
            region in the background image.
        """
        if isinstance(background, (str, int)):
            bg = self[self.__lookup__(background)]
        else:
            bg = background
        if isinstance(bg, ImageFile):
            bg = bg.image
        bg = bg.view(ImageArray)
        bgmask = ~bg.mask
        bgmean = bg[bgmask].mean()
        for i in range(len(self)):
            im_mean = self._stack[:, :, i][bgmask].mean()
            self._stack[:, :, i] -= bg * im_mean / bgmean
        return self


class ImageStack(StackAnalysisMixin, ImageStackMixin, ImageFolderMixin, DiskBasedFolderMixin, baseFolder):
    """An alternative implementation of an image stack based on baseFolder."""



"""Example using select method to pick out data."""

# pylint: disable=invalid-name
from Stoner import Data

d = Data("sample.txt", setas="xy")
d.plot(fmt="b-")
d.select(Temp__gt=75).select(Res__between=(5.3, 6.3)).plot(
    fmt="ro", label="portion 1"
)
d.select(Temp__lt=30).plot(fmt="g<", label="portion 2")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Stoner Package specific exceptions."""


class StonerLoadError(Exception):
    """An exception thrown by the file loading routines in the Stoner Package.

    This special exception is thrown when one of the subclasses of :py:class:`Stoner.Core.DataFile`
    attempts and then fails to load some data from disk. Generally speaking this is not a real
    error, but simply indicates that the file format is not recognised by that particular subclass,
    and thus another subclass should have a go instead.
    """


class StonerUnrecognisedFormat(IOError):
    """An exception thrown by the file loading routines in the Stoner Package.

    This special exception is thrown when none of the subclasses was able to load the specified file.
    """


class StonerSetasError(AttributeError):
    """An exception tjrown when we try to access a column in data without setas being set."""


class StonerAssertionError(RuntimeError):
    """An exception raised when the library thinks an assertion has failed."""


def assertion(condition, message="Library Assertion Error set"):
    """Raise an error when condition is false.

    A utility functiuon to be used when assert might have been."""
    if not condition:
        raise StonerAssertionError(message)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""":py:class:`lmfit.Model` model classes and functions for various thermal physics models."""
# pylint: disable=invalid-name
__all__ = [
    "Arrhenius",
    "ModArrhenius",
    "Model",
    "NDimArrhenius",
    "VFTEquation",
    "arrhenius",
    "modArrhenius",
    "nDimArrhenius",
    "np",
    "vftEquation",
]


import numpy as np
import scipy.constants as consts
from scipy.optimize import curve_fit

from lmfit import Model
from lmfit.models import update_param_vals


def arrhenius(x, A, DE):
    r"""Arrhenius Equation without T dependent prefactor.

    Args:
        x (array): temperatyre data in K
        A (float): Prefactor - temperature independent. See :py:func:modArrhenius for temperaure dependent version.
        DE (float): Energy barrier in *eV*.

    Return:
        Typically a rate corresponding to the given temperature values.

    The Arrhenius function is defined as :math:`\tau=A\exp\left(\frac{-\Delta E}{k_B x}\right)` where
    :math:`k_B` is Boltzmann's constant.

    Example:
        .. plot:: samples/Fitting/Arrhenius.py
            :include-source:
            :outname: arrhenius

    """
    _kb = consts.physical_constants["Boltzmann constant"][0] / consts.physical_constants["elementary charge"][0]
    return A * np.exp(-DE / (_kb * x))


def nDimArrhenius(x, A, DE, n):
    r"""Arrhenius Equation without T dependent prefactor for various dimensions.

    Args:
        x (array): temperatyre data in K
        A (float): Prefactor - temperature independent. See :py:func:modArrhenius for temperaure dependent version.
        DE (float): Energy barrier in *eV*.
        n (float): The dimensionalirty of the model

    Return:
        Typically a rate corresponding to the given temperature values.

    The Arrhenius function is defined as :math:`\tau=A\exp\left(\frac{-\Delta E}{k_B x^n}\right)` where
    :math:`k_B` is Boltzmann's constant.

    Example:
        .. plot:: samples/Fitting/nDimArrhenius.py
            :include-source:
            :outname: nDimarrehenius
    """
    return arrhenius(x**n, A, DE)


def modArrhenius(x, A, DE, n):
    r"""Arrhenius Equation with a variable T power dependent prefactor.

    Args:
        x (array): temperatyre data in K
        A (float): Prefactor - temperature independent. See :py:func:modArrhenius for temperaure dependent version.
        DE (float): Energy barrier in *eV*.
        n (float): The exponent of the temperature pre-factor of the model

    Return:
        Typically a rate corresponding to the given temperature values.

    The modified Arrhenius function is defined as :math:`\tau=Ax^n\exp\left(\frac{-\Delta E}{k_B x}\right)` where
    :math:`k_B` is Boltzmann's constant.

    Example:
        .. plot:: samples/Fitting/modArrhenius.py
            :include-source:
            :outname: modarrhenius
    """
    return (x**n) * arrhenius(x, A, DE)


def vftEquation(x, A, DE, x_0):
    r"""Vogel-Flucher-Tammann (VFT) Equation without T dependent prefactor.

    Args:
        x (float): Temperature in K
        A (float): Prefactror (not temperature dependent)
        DE (float): Energy barrier in eV
        x_0 (float): Offset temperature in K

    Return:
        Rates according the VFT equation.

    The VFT equation is defined as as :math:`\tau = A\exp\left(\frac{DE}{x-x_0}\right)` and represents
    a modified form of the Arrenhius distribution with a freezing point of :math:`x_0`.

    Example:
        .. plot:: samples/Fitting/vftEquation.py
            :include-source:
            :outname: vft
    """
    _kb = consts.physical_constants["Boltzmann constant"][0] / consts.physical_constants["elementary charge"][0]
    X = np.where(np.isclose(x, x_0), 1e-8, x - x_0)
    y = A * np.exp(-DE / (_kb * X))

    return y


class Arrhenius(Model):
    r"""Arrhenius Equation without T dependent prefactor.

    Args:
        x (array): temperatyre data in K
        A (float): Prefactor - temperature independent. See :py:func:modArrhenius for temperaure dependent version.
        DE (float): Energy barrier in *eV*.

    Return:
        Typically a rate corresponding to the given temperature values.

    The Arrhenius function is defined as :math:`\tau=A\exp\left(\frac{-\Delta E}{k_B x}\right)` where
    :math:`k_B` is Boltzmann's constant.

    Example:
        .. plot:: samples/Fitting/Arrhenius.py
            :include-source:
            :outname: arrhenius-class
    """

    display_names = ["A", r"\Delta E"]

    def __init__(self, *args, **kwargs):
        """Configure default function to fit."""
        super().__init__(arrhenius, *args, **kwargs)

    def guess(self, data, x=None, **kwargs):
        """Estimate fitting parameters from data."""
        _kb = consts.physical_constants["Boltzmann constant"][0] / consts.physical_constants["elementary charge"][0]

        d1, d2 = 1.0, 0.0
        if x is not None:
            d1, d2 = np.polyfit(-1.0 / x, np.log(data), 1)
        pars = self.make_params(A=np.exp(d2), DE=_kb * d1)
        return update_param_vals(pars, self.prefix, **kwargs)


class NDimArrhenius(Model):
    r"""Arrhenius Equation without T dependent prefactor for various dimensions.

    Args:
        x (array): temperatyre data in K
        A (float): Prefactor - temperature independent. See :py:func:modArrhenius for temperaure dependent version.
        DE (float): Energy barrier in *eV*.
        n (float): The dimensionalirty of the model

    Return:
        Typically a rate corresponding to the given temperature values.

    The Arrhenius function is defined as :math:`\tau=A\exp\left(\frac{-\Delta E}{k_B x^n}\right)` where
    :math:`k_B` is Boltzmann's constant.

    Example:
        .. plot:: samples/Fitting/nDimArrhenius.py
            :include-source:
            :outname: nDimarrhenius-class
    """

    display_names = ["A", r"\Delta E", "n"]

    def __init__(self, *args, **kwargs):
        """Configure Initial fitting function."""
        super().__init__(nDimArrhenius, *args, **kwargs)

    def guess(self, data, x=None, **kwargs):
        """Guess paramneters from a set of data."""
        _kb = consts.physical_constants["Boltzmann constant"][0] / consts.physical_constants["elementary charge"][0]

        d1, d2 = 1.0, 0.0
        if x is not None:
            d1, d2 = np.polyfit(-1.0 / x, np.log(data), 1)
        pars = self.make_params(A=np.exp(d2), DE=_kb * d1, n=1.0)
        return update_param_vals(pars, self.prefix, **kwargs)


class ModArrhenius(Model):
    r"""Arrhenius Equation with a variable T power dependent prefactor.

    Args:
        x (array): temperatyre data in K
        A (float): Prefactor - temperature independent. See :py:func:modArrhenius for temperaure dependent version.
        DE (float): Energy barrier in *eV*.
        n (float): The exponent of the temperature pre-factor of the model

    Return:
        Typically a rate corresponding to the given temperature values.

    The Arrhenius function is defined as :math:`\tau=Ax^n\exp\left(\frac{-\Delta E}{k_B x}\right)` where
    :math:`k_B` is Boltzmann's constant.

    Example:
        .. plot:: samples/Fitting/modArrhenius.py
            :include-source:
            :outname: modarrhenius-class
    """

    display_names = ["A", r"\Delta E", "n"]

    def __init__(self, *args, **kwargs):
        """Configure Initial fitting function."""
        super().__init__(modArrhenius, *args, **kwargs)

    def guess(self, data, x=None, **kwargs):
        """Guess paramneters from a set of data."""
        _kb = consts.physical_constants["Boltzmann constant"][0] / consts.physical_constants["elementary charge"][0]

        d1, d2 = 1.0, 0.0
        if x is not None:
            d1, d2 = np.polyfit(-1.0 / x, np.log(data / x), 1)
        pars = self.make_params(A=np.exp(d2), DE=_kb * d1, n=1.0)
        return update_param_vals(pars, self.prefix, **kwargs)


class VFTEquation(Model):
    r"""Vogel-Flucher-Tammann (VFT) Equation without T dependent prefactor.

    Args:
        x (array): Temperature in K
        A (float): Prefactror (not temperature dependent)
        DE (float): Energy barrier in eV
        x_0 (float): Offset temperature in K

    Return:
        Rates according the VFT equation.

    The VFT equation is defined as as :math:`\tau = A\exp\left(\frac{DE}{x-x_0}\right)` and represents
    a modified form of the Arrenhius distribution with a freezing point of :math:`x_0`.

    See :py:func:`vftEquation` for an example.

    Example:
        .. plot:: samples/Fitting/vftEquation.py
            :include-source:
            :outname: vft-class
    """

    display_names = ["A", r"\Delta E", "x_0"]

    nan_policy = "omit"

    def __init__(self, *args, **kwargs):
        """Configure Initial fitting function."""
        super().__init__(vftEquation, *args, **kwargs)

    def guess(self, data, x=None, **kwargs):
        """Guess paramneters from a set of data."""
        _kb = consts.physical_constants["Boltzmann constant"][0] / consts.physical_constants["elementary charge"][0]
        d1, d2, x0 = 1.0, 0.0, 1.0
        yy = np.log(data)
        if x is not None:
            # Getting a good x_0 is critical, so we first of all use poly fit to look
            x0 = x[np.argmin(np.abs(data))] * 0.95

            def _find_x0(x, d1, d2, x0):
                X = np.where(np.isclose(x, x0), 1e-8, x - x0)
                y = d2 - (d1 / X)
                return y

            popt = curve_fit(_find_x0, x, yy, p0=[1.0 / _kb, 25, x0])[0]
            d1, d2, x0 = popt
        pars = self.make_params(A=np.exp(d2), DE=_kb * d1, x_0=x0)
        return update_param_vals(pars, self.prefix, **kwargs)



"""Example of Arrhenius Fit."""

# pylint: disable=invalid-name
from numpy import logspace, log10
from numpy.random import normal

from Stoner import Data
from Stoner.analysis.fitting.models.thermal import vftEquation, VFTEquation

# Make some data
T = logspace(log10(200), log10(350), 51)
params = (1e16, 0.5, 150)
noise = 0.5
R = vftEquation(T, *params) * normal(size=len(T), scale=noise, loc=1.0)
dR = vftEquation(T, *params) * noise
d = Data(T, R, dR, setas="xy.", column_headers=["T", "Rate"])

# Plot the data points.
d.plot(fmt="r.", label="Data Points")

# Turn on the sigma column (error bars look messy on plot due to logscale)
d.setas[2] = "e"

# Curve_fit on its own
d.curve_fit(vftEquation, p0=params, result=True, header="curve_fit")

# lmfit uses some guesses
p0 = params
d.lmfit(VFTEquation, result=True, header="lmfit")

# Plot these results too
d.setas = "x..yy"
d.plot(fmt=["b-", "g-"])
# Annotate the graph
d.annotate_fit(
    vftEquation,
    x=0.25,
    y=0.35,
    fontdict={"size": "x-small", "color": "blue"},
    mode="eng",
)
d.annotate_fit(
    VFTEquation,
    x=0.5,
    y=0.35,
    prefix="VFTEquation",
    fontdict={"size": "x-small", "color": "green"},
    mode="eng",
)

# reset the columns for the fit
d.setas = "xye.."
# Now do the odr fit (will overwrite lmfit's metadata)
d.odr(VFTEquation, result=True)
d.setas = "x4.y"

# And plot and annotate
d.plot(fmt="m-", label="Orthogonal distance")
d.annotate_fit(
    VFTEquation,
    x=0.75,
    y=0.35,
    fontdict={"size": "x-small", "color": "magenta"},
    mode="eng",
)

# Finally tidy up the plot a bit
d.yscale = "log"
d.ylim = (1e-35, 1e10)
d.title = "VFT Equation Test Fit"
d.ylabel = "Rate"
d.xlabel = "Temperature (K)"



# -*- coding: utf-8 -*-
"""Demonstrate ImageFolder mask_select."""

import threading
import time

import numpy as np
from matplotlib import pyplot as plt

from Stoner import ImageFolder, __datapath__
from Stoner.HDF5 import STXMImage
from Stoner.Image.widgets import send_event as _event


def fake_user_action(image):
    """Send events to the selection widget to simulate a user."""
    time.sleep(1)
    _event(
        image,
        ["draw", "on_click"],
        xdata=50,
        ydata=20,
        button=1,
        dblclick=False,
    )
    for y in np.linspace(20, 80, 10):
        _event(image, "draw", xdata=50, ydata=y)
        time.sleep(0.1)
    _event(image, "keypress", xdata=50, ydata=80, key="c")
    time.sleep(0.5)
    _event(image, "keypress", xdata=50, ydata=80, key="i")
    time.sleep(0.5)
    _event(
        image,
        ["draw", "keypress"],
        xdata=50,
        ydata=80,
        button=1,
        dblclick=False,
        key="enter",
    )


fldr = ImageFolder(
    __datapath__, pattern="Sample*.hdf5", type=STXMImage, recursive=False
)

# Start the scripted control
fake_user = threading.Thread(target=fake_user_action, args=(fldr[0],))
fake_user.start()

fldr.mask_select()
for i in range(4):
    fldr += fldr[0]
    fldr += fldr[1]
fig = plt.figure(figsize=(8, 4))
fldr.montage(figure=fig, plots_per_page=4)



# -*- coding: utf-8 -*-
"""Test Stoner.core.exceptions module"""
import pytest
from Stoner.core.exceptions import assertion


def test_assertion():
    with pytest.raises(RuntimeError):
        assertion(False, "Triggered an assertion")


if __name__ == "__main__":
    pytest.main(["--pdb", __file__])



"""Simple plotting with a template."""

# pylint: disable=invalid-name, no-member
from cycler import cycler

from Stoner import Data
from Stoner.plot.formats import DefaultPlotStyle

p = Data("sample.txt", setas="xy", template=DefaultPlotStyle())
p.template(  # pylint: disable=not-callable
    axes__prop_cycle=cycler("color", ["r", "g", "b"])
)  # pylint: disable=not-callable
p.plot()
p.y += 1.0
p.plot()
p.y += 1.0
p.plot()



"""Test Weak-localisation fitting."""

# pylint: disable=invalid-name
from copy import copy

from numpy import linspace, ones_like
from numpy.random import normal

from Stoner import Data
from Stoner.analysis.fitting.models.e_transport import WLfit, wlfit

B = linspace(-8, 8, 201)
params = [1e-3, 2.0, 0.25, 1.4]
G = wlfit(B, *params) + normal(size=len(B), scale=5e-7)
dG = ones_like(B) * 5e-7
d = Data(
    B,
    G,
    dG,
    setas="xye",
    column_headers=["Field $\\mu_0H (T)$", "Conductance", "dConductance"],
)

d.curve_fit(wlfit, p0=copy(params), result=True, header="curve_fit")

d.setas = "xye"
d.lmfit(WLfit, result=True, header="lmfit")

d.setas = "xyeyy"
d.plot(fmt=["r.", "b-", "g-"])

d.annotate_fit(
    wlfit, x=0.05, y=0.75, fontdict={"size": "x-small", "color": "blue"}
)
d.annotate_fit(
    WLfit,
    x=0.05,
    y=0.5,
    fontdict={"size": "x-small", "color": "green"},
    prefix="WLfit",
)
d.title = "Weak Localisation Fit"



"""Test langevin fitting."""

# pylint: disable=invalid-name
from copy import copy

from numpy import linspace, ones_like
from numpy.random import normal

from Stoner import Data
from Stoner.analysis.fitting.models.magnetism import langevin, Langevin

B = linspace(-0.01, 0.01, 100)
params = [1, 1.0e-11, 250]
G = langevin(B, *params) + normal(size=len(B), scale=5e-3)
dG = ones_like(B) * 5e-3
d = Data(
    B,
    G,
    dG,
    setas="xye",
    column_headers=["Field $\\mu_0H (T)$", "Moment", "dM"],
)

func = lambda H, M_s, m: langevin(H, M_s, m, 250)

d.curve_fit(func, p0=copy(params)[0:2], result=True, header="curve_fit")

d.setas = "xye"
fit = Langevin()
fit.params = fit.guess(G, x=B)
fit.params["T"].value = 250
fit.params["T"].vary = False

d.lmfit(fit, p0=fit.params, result=True, header="lmfit")


d.setas = "xyeyy"
d.plot(fmt=["r.", "b-", "g-"])

d.annotate_fit(
    func,
    x=0.1,
    y=0.5,
    fontdict={"size": "x-small", "color": "blue"},
    mode="eng",
)
d.annotate_fit(
    Langevin,
    x=0.1,
    y=0.25,
    fontdict={"size": "x-small", "color": "green"},
    prefix="Langevin",
    mode="eng",
)
d.title = "langevin Fit"



"""Plot data defined on a matrix."""

# pylint: disable=invalid-name
import numpy as np

from Stoner import Data

x, y = np.meshgrid(np.linspace(-2, 2, 101), np.linspace(-2, 2, 101))
z = np.cos(4 * np.pi * np.sqrt(x**2 + y**2)) * np.exp(-np.sqrt(x**2 + y**2))

p = Data()
p = p & np.linspace(-2, 2, 101) & z
p.column_headers = ["X"]
for i, v in enumerate(np.linspace(-2, 2, 101)):
    p.column_headers[i + 1] = str(v)

p.plot_matrix(xlabel="x", ylabel="y", title="Data as Matrix")



"""VSM Analysis Code example v2.

Author: Rowan Temple.     Date:11/2011

***YOU CANNOT RUN THIS FILE WITHIN IDLE! You must run it in a command window. (matplotlib interactive
thing gets mixed up from within IDLE.)***

Editing VSM files (taking off drift, setting middle to zero etc.)
When entering path names you need the folder structure up to the point where this program is.
I think there will be difficulties if this program is run in ipython to do with graph not displaying,
easily fixed

v2 corrections: less automation saving time
                graph is used in interactive mode
                shift algorithm is different to make for better zeroing

"""

# pylint: disable=invalid-name, redefined-outer-name
import os

import numpy as np
import matplotlib.pyplot as plot

import Stoner

plot.ion()  # put the plotting in interactive mode
raw_input = input
print("Analysis program for VSM files\n\n\n")
print("Please wait while program loads...\n")
# VSMcalibration=0.86224   #m to X ratio, determined by reference Py sample


def deleteCorruptLines(data):
    """Take data array and returns data with lines with 6:0 or --- in them deleted."""
    delCount = 0
    for line in data:
        if line.find("6:0") != -1:
            data.pop(data.index(line))
            delCount += 1
        if line.find("---") != -1:
            data.pop(data.index(line))
            delCount += 1
    if delCount > 6:
        input(
            "I've detected a lot of bad data in your file you should check it"
        )
    return data


def driftEliminator(data, N):
    """Take off a linear drift of m with time from the data."""
    maxHarg = int(np.argmax(data[: len(data[:, 0]) / 2, 1]))
    finalHarg = int(len(data[:, 1]) - 1)
    # linear fit to saturated data (+- N points from max field and final field)
    finalFit = np.polyfit(
        data[finalHarg - N : finalHarg, 1],
        data[finalHarg - N : finalHarg, 2],
        1,
    )
    firstFit = np.polyfit(
        data[maxHarg - N : maxHarg + N, 1],
        data[maxHarg - N : maxHarg + N, 2],
        1,
    )
    pt = np.zeros((2, 2))
    pt[0] = [
        data[maxHarg, 0],
        firstFit[0] * data[maxHarg, 1] + firstFit[1],
    ]  # (time,m) at max field
    pt[1] = [data[finalHarg, 0], finalFit[0] * data[finalHarg, 1] + finalFit[1]]
    # Delete drift
    for i in range(len(data[:, 0])):
        data[i, 2] = data[i, 2] - (pt[1, 1] - pt[0, 1]) / (
            pt[1, 0] - pt[0, 0]
        ) * (data[i, 0] - pt[0, 0])
    return data


def shift(data, N):
    """Translate curve in y so that halfway between the saturated fieldsis zero.

    (uses the average y value of all the saturated points given to determine upper and lower bounds of curve
    """
    maxHarg = int(
        np.argmax(data[: len(data[:, 0]) / 2, 1])
    )  # gives the data index of maximum field (in the first half of the data
    minHarg = int(np.argmin(data[:, 1]))
    # average saturated data (+- N points from max/min field)
    highAve = np.average(data[(maxHarg - N) : (maxHarg + N), 2])
    lowAve = np.average(data[minHarg - N : minHarg + N, 2])
    data[:, 2] = data[:, 2] - (highAve + lowAve) / 2.0
    return data


def diamagBackgroundRem(data, N):
    """Remove a diamagnetic background using a linear fit to the N data points surrounding the max and min field."""
    maxHarg = int(np.argmax(data[: len(data[:, 0]) / 2, 1]))
    minHarg = int(np.argmin(data[:, 1]))

    # linear fit to saturated data (+- N points from max/min field)
    highFit = np.polyfit(
        data[maxHarg - N : maxHarg + N, 1],
        data[maxHarg - N : maxHarg + N, 2],
        1,
    )
    lowFit = np.polyfit(
        data[minHarg - N : minHarg + N, 1],
        data[minHarg - N : minHarg + N, 2],
        1,
    )
    # Average grad
    fitGrad = (highFit[0] + lowFit[0]) / 2
    # Delete linear grad from all data
    data[:, 2] = data[:, 2] - (fitGrad * data[:, 1])
    return data


def invert(Data):
    """Flip data in y axis."""
    for i in range(len(Data.data[:, "0"])):
        Data.data[i, Data.find_col("m (emu)")] = -Data.data[
            i, Data.find_col("m (emu)")
        ]
    return Data


def makeTruem(Data):
    """VSM takes m from lock in X, this takes m from lock in R, useful if theta!=0."""
    VSMcalibration = (
        Data.data[5, Data.find_col("m (emu)")]
        / Data.data[5, Data.find_col("X (V)")]
    )  # taken from row 5 at random
    for i in range(len(Data.data[:, "0"])):
        Data.data[i, Data.find_col("m (emu)")] = (
            VSMcalibration * Data.data[i, Data.find_col("R (V)")]
        )
    return Data


def plotmH(data):
    """Take a stoner type data source."""
    plot.clf()
    plot.xlabel("H(T)")
    plot.ylabel("m(1e-5 emu)")
    plot.plot(data.column("H"), data.column("m (emu)"), "b-")
    plot.draw()


def splitFileName(myFileName):
    """Split a file name into its name and its extension part.

    (returns two part list ['name','ext'] or ['name',''] if no extension"""
    for i in range(len(myFileName) - 1, -1, -1):
        if myFileName[i] == ".":
            fileName = myFileName[:i]
            fileExt = myFileName[i:]
            return [fileName, fileExt]
    return [myFileName, ""]


def editData(Data, operations):
    """Take stoner type Data file and an operations list and performs the operations listed."""
    if 0 in operations:
        return Data
    N = int(input("Input the number of saturated data points on each arm:   "))
    if 1 in operations:
        Data = makeTruem(Data)
    if 2 in operations:
        Data.data = driftEliminator(Data.data, N)
    if 3 in operations:
        Data.data = diamagBackgroundRem(Data.data, N)
    if 4 in operations:
        Data.data = shift(Data.data, N)
    if 5 in operations:
        Data = invert(Data)
    return Data


# Set up a directory and determine files to be processed
directoryName = input("Enter path to directory in which files are stored:   ")
os.chdir(directoryName)
filenames = os.listdir(directoryName)
i = 0
for item in filenames:
    print(i, ". ", item)
    i += 1
fCounter = int(
    input(
        "Enter the number of the file you wish to start at (program will cycle through files from that point.):\n"
    )
)
if "EditedFiles" not in filenames:
    os.mkdir("EditedFiles")

timeout = 0
# Main program loop, cycle through selected files
while True:
    path = filenames[fCounter]
    fCounter += 1
    fr = open(path, "r", encoding="utf-8")
    data = fr.readlines()  # Get the file into an array
    fr.close()
    pathsplit = splitFileName(path)
    fw = open(
        "EditedFiles/" + pathsplit[0] + "_edit.txt", "w", enconcoding="utf-8"
    )  # create a new file for the edited data
    data = deleteCorruptLines(data)
    fw.writelines(
        data
    )  # put the uncorrupted lines into the file so that Stoner can open it
    fw.close()
    while True:  # open the file
        try:
            Data = Stoner.Data("EditedFiles/" + pathsplit[0] + "_edit.txt")
            break
        except ValueError:
            try:
                Data = Stoner.Data("EditedFiles/" + pathsplit[0] + "_edit.txt")
                break
            except ValueError:
                timeout += 1  # if get 5 files unreadable in a row then finish the program
                print("Could not read file ", path)
                if timeout <= 5:
                    break
    fw.close()
    while True:
        plotmH(Data)
        if not ("Original m (emu)" in Data.column_headers):
            Data.add_column(Data.column("m (emu)"), "Original m (emu)")
        print(
            "\nOK you have 5 options here, have a look at the plot and please tell me which ones ",
            'you would like by entering a string eg "125" for options 1, 2 and 5\n',
        )
        print(
            "0. Do nothing \n",
            "1.  Give true m from R (if theta is not zeroed in your data this will give you true m) \n",
            "2.  Remove a drift in m with time \n",
            "3.  Remove a diamagnetic background \n",
            "4.  Remove an overall shift in m \n",
            "5.  Reflect graph in y axis \n",
        )

        strOp = input("")
        operations = []  # array of options selected
        for i in range(len(strOp.strip())):
            operations.append(int(strOp[i]))
        t = Data.clone  # edit a copied array.
        t = editData(t, operations)
        plotmH(t)
        whatNext = input(
            "Press enter to save changes, r to restart or q to quit the program:  "
        )
        if whatNext == "r":
            continue
        if whatNext == "q":
            break
        Data = t
        Data.save(
            "EditedFiles/" + pathsplit[0] + "_edit.txt"
        )  # overwrite the file created earlier
        break
    if (
        whatNext == "q"
        or input(
            "Press enter to do file {} or q to quit:".format(
                filenames[fCounter]
            )
        )
        == "q"
    ):
        break
plot.close()
plot.ioff()



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""Support functions for Stoner package.

These functions depend only on Stoner.compat which is used to ensure a consistent namespace between python 2.7 and 3.x.
"""
__all__ = [
    "AttributeStore",
    "all_size",
    "all_type",
    "fix_signature",
    "format_error",
    "format_val",
    "html_escape",
    "isanynone",
    "isComparable",
    "isnone",
    "isiterable",
    "isLikeList",
    "isproperty",
    "isTuple",
    "copy_into",
    "make_Data",
    "quantize",
    "tex_escape",
    "typedList",
    "Options",
    "get_option",
    "set_option",
    "ordinal",
    "file",
    "decorators",
]
from collections.abc import Iterable, MutableSequence
import inspect
from copy import deepcopy
from importlib import import_module

from numpy import log10, floor, logical_and, isnan, round, ndarray, dtype  # pylint: disable=redefined-builtin


from ..compat import bytes2str
from .classes import attributeStore as AttributeStore, typedList, Options, get_option, set_option, copy_into
from .tests import all_size, all_type, isanynone, isComparable, isiterable, isLikeList, isnone, isproperty, isTuple
from .formatting import format_error, format_val, quantize, html_escape, tex_escape, ordinal
from . import decorators
from .decorators import make_Data, fix_signature

operator = {
    "eq": lambda k, v: k == v,
    "ne": lambda k, v: k != v,
    "contains": lambda k, v: v in k,
    "in": lambda k, v: k in v,
    "icontains": lambda k, v: k.upper() in str(v).upper(),
    "iin": lambda k, v: str(v).upper() in k.upper(),
    "lt": lambda k, v: k < v,
    "le": lambda k, v: k <= v,
    "gt": lambda k, v: k > v,
    "ge": lambda k, v: k >= v,
    "between": lambda k, v: logical_and(min(v) < k, k < max(v)),
    "ibetween": lambda k, v: logical_and(min(v) <= k, k <= max(v)),
    "ilbetween": lambda k, v: logical_and(min(v) <= k, k < max(v)),
    "iubetween": lambda k, v: logical_and(min(v) < k, k <= max(v)),
    "startswith": lambda k, v: str(v).startswith(k),
    "istartswith": lambda k, v: str(v).upper().startswith(k.upper()),
    "endsswith": lambda k, v: str(v).endswith(k),
    "iendsswith": lambda k, v: str(v).upper().endswith(k.upper()),
}



# -*- coding: utf-8 -*-
"""Test Util module"""


import sys
import os.path as path
import Stoner.Util as SU
from Stoner import Data
import pytest

pth = path.dirname(__file__)
pth = path.realpath(path.join(pth, "../../"))
sys.path.insert(0, pth)

def is_2tuple(x):
    """Return tru if x is a length two tuple of floats."""
    return isinstance(x, tuple) and len(x) == 2 and isinstance(x[0], float) and isinstance(x[1], float)


@pytest.mark.parametrize("meth", ["linear_intercept", "susceptibility", "delta_M"])
def test_hysteresis(meth):
    """Test the hysteresis analysis code."""
    x = SU.hysteresis_correct(
        path.join(pth, "./sample-data/QD-SQUID-VSM.dat"), setas="3.xy", h_sat_method=meth, saturated_fraction=0.25
    )
    assert (
        "Hc" in x and "Area" in x and "Hsat" in x and "BH_Max" in x and "BH_Max_H" in x
    ), "Hystersis loop analysis keys not present."

    assert is_2tuple(x["Hc"]) and x["Hc"][0] + 578 < 1.0, "Failed to find correct Hc in a SQUID loop"
    assert isinstance(x["Area"], float) and -0.0137 < x["Area"] < -0.0136, "Incorrect calculation of area under loop"


def test_failures():
    with pytest.raises(ValueError):
        _ = SU.hysteresis_correct(
            path.join(pth, "./sample-data/QD-SQUID-VSM.dat"),
            setas="3.xy",
            h_sat_method="bad",
            saturated_fraction=0.25,
        )


def test_split_up_down():
    testd = Data(path.join(pth, "./sample-data/QD-SQUID-VSM.dat"), setas="3.xy")
    fldr = SU.split_up_down(testd)
    x = SU.hysteresis_correct(fldr["falling"][0], saturated_fraction=0.25)
    assert (
        "Hc" in x and "Area" in x and "Hsat" in x and "BH_Max" in x and "BH_Max_H" in x
    ), "Hystersis loop analysis keys not present."

    assert x["Hc_mean"] - 570 < 1.0, "Failed to find correct Hc in a SQUID loop"
    assert isinstance(x["Area"], float) and -0.0055 < x["Area"] < -0.0054, "Incorrect calculation of area under loop"


if __name__ == "__main__":  # Run some tests manually to allow debugging
    pytest.main(["--pdb", __file__])



# -*- coding: utf-8 -*-
"""Functions for formatting data."""

__all__ = ["format_error", "format_val", "quantize", "tex_escape", "ordinal"]

import re
from typing import Optional, Any
from html import escape as html_escape
from numpy import log10, floor, abs, isnan, round  # pylint: disable=redefined-builtin
from ..compat import bytes2str
from ..core.Typing import Numeric, NumericArray

prefs = {
    "text": {
        3: "k",
        6: "M",
        9: "G",
        12: "T",
        15: "P",
        18: "E",
        21: "Z",
        24: "Y",
        -3: "m",
        -6: "u",
        -9: "n",
        -12: "p",
        -15: "f",
        -18: "a",
        -21: "z",
        -24: "y",
    },
    "latex": {
        3: "k",
        6: "M",
        9: "G",
        12: "T",
        15: "P",
        18: "E",
        21: "Z",
        24: "Y",
        -3: "m",
        -6: r"\mu",
        -9: "n",
        -12: "p",
        -15: "f",
        -18: "a",
        -21: "z",
        -24: "y",
    },
    "html": {
        3: "k",
        6: "M",
        9: "G",
        12: "T",
        15: "P",
        18: "E",
        21: "Z",
        24: "Y",
        -3: "m",
        -6: r"&micro;",
        -9: "n",
        -12: "p",
        -15: "f",
        -18: "a",
        -21: "z",
        -24: "y",
    },
}


def format_error(value: Numeric, error: Optional[Numeric] = None, **kargs: Any) -> str:
    r"""Format answer with the uncertaintly to 1sf and the value to no more sf's than the uncertainty.

    Args:
        value (float):
            The value to be formatted
        error (float):
            The uncertainty in the value

    Keyword Arguments:
        fmt (str):
            Specify the output format, options are:
                -  "text" - plain text output
                - "latex" - latex output
                - "html" - html entities
        escape (bool):
            Specifies whether to escape the prefix and units for unprintable characters in non text formats (default
            False)
        mode (string):
            If "float" (default) the number is formatted as is, if "eng" the value and error is converted
            to the next samllest power of 1000 and the appropriate SI index appended. If mode is "sci" then a
            scientific, i.e. mantissa and exponent format is used.
        units (string):
            A suffix providing the units of the value. If si mode is used, then appropriate si prefixes are
            prepended to the units string. In LaTeX mode, the units string is embedded in \mathrm
        prefix (string):
            A prefix string that should be included before the value and error string. in LaTeX mode this is
            inside the math-mode markers, but not embedded in \mathrm.

    Returns:
        String containing the formatted number with the eorr to one s.f. and value to no more d.p. than the error.
    """
    mode = kargs.get("mode", "float")
    units = kargs.get("units", "")
    prefix = kargs.get("prefix", "")
    latex = kargs.get("latex", False)
    fmt = kargs.get("fmt", "latex" if latex else "text")
    escape = kargs.get("escape", False)
    escape_func = {"latex": tex_escape, "html": html_escape}.get(fmt, lambda x: x)

    if error == 0.0 or isnan(error):  # special case for zero uncertainty
        return format_val(value, **kargs)

    if escape:
        prefix = escape_func(prefix)
        units = escape_func(units)

    # Sort out special formatting for different modes
    if mode == "float":  # Standard
        suffix_val = ""
    elif mode == "eng":  # Use SI prefixes
        if -1 <= floor(log10(abs(value))) <= 3:
            v_mag = 0
        else:
            v_mag = floor(log10(abs(value)) / 3.0) * 3.0
        prefixes = prefs.get(fmt, prefs["text"])
        if v_mag in prefixes:
            if fmt == "latex":
                suffix_val = rf"\,\mathrm{{{{{prefixes[v_mag]}}}}}"
            else:
                suffix_val = " " + prefixes[v_mag]
            value /= 10**v_mag
            error /= 10**v_mag
        else:  # Implies 10^-3<x<10^3
            suffix_val = ""
    elif mode == "sci":  # Scientific mode - raise to common power of 10
        v_mag = floor(log10(abs(value)))
        if fmt == "latex":
            suffix_val = rf"\times 10^{{{{{int(v_mag)}}}}}\,"
        elif fmt == "html":
            suffix_val = f"&times; 10<sup>{int(v_mag)}</sup> "
        else:
            suffix_val = f"E{int(v_mag)} "
        value /= 10**v_mag
        error /= 10**v_mag
    else:  # Bad mode
        raise RuntimeError(f"Unrecognised mode: {mode} in format_error")

    # Now do the rounding of the value based on error to 1 s.f.
    e2 = error
    u_mag = floor(log10(abs(error)))  # work out the scale of the error
    error = round(error / 10**u_mag) * 10**u_mag  # round the error, but this could round to 0.x0
    u_mag = floor(log10(error))  # so go round the loop again
    error = round(e2 / 10**u_mag) * 10**u_mag  # and get a new error magnitude
    value = round(value / 10**u_mag) * 10**u_mag
    u_mag = min(0, u_mag)  # Force integer results to have no dp

    # Protect {} in units string
    units = units.replace("{", "{{").replace("}", "}}")
    prefix = prefix.replace("{", "{{").replace("}", "}}")
    if fmt == "latex":  # Switch to latex math mode symbols
        val_fmt_str = rf"${prefix}{{:.{int(abs(u_mag))}f}}\pm "
        if units != "":
            suffix_fmt = rf"\mathrm{{{{{units}}}}}"
        else:
            suffix_fmt = ""
        suffix_fmt += "$"
    elif fmt == "html":  # Switch to latex math mode symbols
        val_fmt_str = rf"{prefix}{{:.{int(abs(u_mag))}f}}&plusmin;"
        suffix_fmt = units
    else:  # Plain text
        val_fmt_str = rf"{prefix}{{:.{int(abs(u_mag))}f}}+/-"
        suffix_fmt = units
    if u_mag < 0:  # the error is less than 1, so con strain decimal places
        err_fmt_str = r"{:." + str(int(abs(u_mag))) + "f}"
    else:  # We'll be converting it to an integer anyway
        err_fmt_str = r"{}"
    fmt_str = val_fmt_str + err_fmt_str + suffix_val + suffix_fmt
    if error >= 1.0:
        error = int(error)
        value = int(value)
    return fmt_str.format(value, error)


def format_val(value: Numeric, **kargs: Any) -> str:
    r"""Format a number as an SI quantity.

    Args:
        value(float):
            Value to format

    Keyword Arguments:
        fmt (str):
            Specify the output format, options are:
                -  "text" - plain text output
                - "latex" - latex output
                - "html" - html entities
        escape (bool):
            Specifies whether to escape the prefix and units for unprintable characters in non text formats (default
            False)
        mode (string):
            If "float" (default) the number is formatted as is, if "eng" the value and error is converted
            to the next samllest power of 1000 and the appropriate SI index appended. If mode is "sci" then a
            scientific, i.e. mantissa and exponent format is used.
        units (string):
            A suffix providing the units of the value. If si mode is used, then appropriate si prefixes are
            prepended to the units string. In LaTeX mode, the units string is embedded in \mathrm
        prefix (string):
            A prefix string that should be included before the value and error string. in LaTeX mode this is
            inside the math-mode markers, but not embedded in \mathrm.

    Returns:
        (str): The formatted value.
    """
    mode = kargs.pop("mode", "float")
    units = kargs.pop("units", "")
    prefix = kargs.pop("prefix", "")
    latex = kargs.pop("latex", False)
    fmt = kargs.pop("fmt", "latex" if latex else "text")
    places = kargs.pop("places", False)
    escape = kargs.pop("escape", False)
    escape_func = {"latex": tex_escape, "html": html_escape}.get(fmt, lambda x: x)

    if escape:
        prefix = escape_func(prefix)
        units = escape_func(units)

    if mode == "float":  # Standard
        suffix_val = ""
        v_mag = floor(log10(abs(value)))
    elif mode == "eng":  # Use SI prefixes
        v_mag = floor(log10(abs(value)) / 3.0) * 3.0
        prefixes = prefs.get(fmt, prefs["text"])
        if v_mag in prefixes:
            if fmt == "latex":
                suffix_val = rf"\mathrm{{{{{prefixes[v_mag]}}}}}"
            else:
                suffix_val = prefixes[v_mag]
            value /= 10**v_mag
        else:  # Implies 10^-3<x<10^3
            suffix_val = ""
    elif mode == "sci":  # Scientific mode - raise to common power of 10
        v_mag = floor(log10(abs(value)))
        if fmt == "latex":
            suffix_val = rf"\times 10^{{{{{int(v_mag)}}}}}"
        elif fmt == "html":
            suffix_val = f"&times; 10<sup>{int(v_mag)}</sup> "
        else:
            suffix_val = f"E{int(v_mag)} "
        value /= 10**v_mag
    else:  # Bad mode
        raise RuntimeError(f"Unrecognised mode: {mode} in format_error")

    # Protect {} in units string
    units = units.replace("{", "{{").replace("}", "}}")
    prefix = prefix.replace("{", "{{").replace("}", "}}")
    if fmt == "latex":  # Switch to latex math mode symbols
        val_fmt_str = rf"${prefix}{{}}"
        if units != "":
            suffix_fmt = rf"\mathrm{{{{{units}}}}}"
        else:
            suffix_fmt = ""
        suffix_fmt += "$"
    elif fmt == "html":  # Switch to latex math mode symbols
        val_fmt_str = rf"{prefix}{{}}"
        suffix_fmt = units
    else:  # Plain text
        val_fmt_str = rf"{prefix}{{}}"
        suffix_fmt = units
    fmt_str = val_fmt_str + suffix_val + suffix_fmt
    if places:
        value = round(value, places)
    ret = fmt_str.format(value)
    return ret


def quantize(number: NumericArray, quantum: Numeric) -> NumericArray:
    """Round a number to the nearest multiple of a quantum.

    Args:
        number (float,array):
            Number(s) to be rounded to the nearest qyuantum
        quantum (float):
            Quantum to round to
    Returns:
        number rounded to qunatum
    """
    return round(number / quantum) * quantum


def tex_escape(text: str) -> str:
    """Escapes spacecial text characters in a string.

    Parameters:
        text (str):
            a plain text message

    Returns:
        the message escaped to appear correctly in LaTeX

    From `Stackoverflow <http://stackoverflow.com/questions/16259923/
    how-can-i-escape-latex-special-characters-inside-django-templates>`
    """
    conv = {
        "&": r"\&",
        "%": r"\%",
        "$": r"\$",
        "#": r"\#",
        "_": r"\_",
        "{": r"\{",
        "}": r"\}",
        "~": r"\textasciitilde{}",
        "^": r"\^{}",
        "\\": r"\textbackslash{}",
        "<": r"\textless",
        ">": r"\textgreater",
    }
    regex = re.compile("|".join(re.escape(bytes2str(key)) for key in sorted(conv.keys(), key=lambda item: -len(item))))
    return regex.sub(lambda match: conv[match.group()], text)


def ordinal(value: int) -> str:
    """Format an integer into an ordinal string.

    Args:
        value (int):
            Number to be written as an ordinal string

    Return:
        (str):
            Ordinal String such as '1st','2nd' etc.
    """
    if not isinstance(value, int):
        raise ValueError

    last_digit = value % 10
    if value % 100 in [11, 12, 13]:
        suffix = "th"
    else:
        suffix = ["th", "st", "nd", "rd", "th", "th", "th", "th", "th", "th"][last_digit]

    return "{}{}".format(value, suffix)



# -*- coding: utf-8 -*-
"""Code adapted from skimage module."""

from __future__ import division

__all__ = ["sign_loss", "prec_loss", "dtype_range", "_dtype", "_dtype2"]
from warnings import warn

import numpy as np

from ..tools.classes import Options

dtype_range = {
    np.bool_: (False, True),
    np.uint8: (0, 255),
    np.uint16: (0, 65535),
    np.int8: (-128, 127),
    np.int16: (-32768, 32767),
    np.int64: (-(2**63), 2**63 - 1),
    np.uint64: (0, 2**64 - 1),
    np.int32: (-(2**31), 2**31 - 1),
    np.uint32: (0, 2**32 - 1),
    np.float32: (-1.0, 1.0),
    np.float64: (-1.0, 1.0),
}

if float(np.version.version.split(".")[1]) < 1.24:
    dtype_range[np.bool8] = (False, True)

integer_types = (np.uint8, np.uint16, np.int8, np.int16)

_supported_types = (
    np.bool_,
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
    np.int8,
    np.int16,
    np.int32,
    np.int64,
    np.float32,
    np.float64,
)

dtype_range[np.float16] = (-1, 1)
_supported_types += (np.float16,)


def sign_loss(dtypeobj_in, dtypeobj):
    """Warn over loss of sign information when converting image."""
    if Options().warnings:
        warn(
            "Possible sign loss when converting negative image of type "
            "%s to positive image of type %s." % (dtypeobj_in, dtypeobj)
        )


def prec_loss(dtypeobj_in, dtypeobj):
    """Warn over precision loss when converting image."""
    if Options().warnings:
        warn("Possible precision loss when converting from " "%s to %s" % (dtypeobj_in, dtypeobj))


def _dtype(itemsize, *dtypes):
    """Return first of `dtypes` with itemsize greater than `itemsize."""
    try:
        ret = next(dt for dt in dtypes if itemsize < np.dtype(dt).itemsize)
    except StopIteration:
        ret = dtypes[0]
    return ret


def _dtype2(kind, bits, itemsize=1):
    """Return dtype of `kind` that can store a `bits` wide unsigned int."""
    c = lambda x, y: x <= y if kind == "u" else x < y
    s = next(i for i in (itemsize,) + (2, 4, 8) if c(bits, i * 8))
    return np.dtype(kind + str(s))


def _scale(a, n, m, dtypeobj_in, dtypeobj, copy=True):
    """Scaleunsigned/positive integers from n to m bits.

    Numbers can be represented exactly only if m is a multiple of n
    Output array is of same kind as input."""
    kind = a.dtype.kind
    if n > m and a.max() <= 2**m:
        mnew = int(np.ceil(m / 2) * 2)
        if mnew > m:
            dtype = "int%s" % mnew
        else:
            dtype = "uint%s" % mnew
        n = int(np.ceil(n / 2) * 2)
        msg = "Downcasting %s to %s without scaling because max " "value %s fits in %s" % (
            a.dtype,
            dtype,
            a.max(),
            dtype,
        )
        if Options().warnings:
            warn(msg)
        return a.astype(_dtype2(kind, m))
    if n == m:
        return a.copy() if copy else a
    if n > m:
        # downscale with precision loss
        prec_loss(dtypeobj_in, dtypeobj)
        if copy:
            b = np.empty(a.shape, _dtype2(kind, m))
            np.floor_divide(a, 2 ** (n - m), out=b, dtype=a.dtype, casting="unsafe")
            return b
        a //= 2 ** (n - m)
        return a
    if m % n == 0:
        # exact upscale to a multiple of n bits
        if copy:
            b = np.empty(a.shape, _dtype2(kind, m))
            np.multiply(a, (2**m - 1) // (2**n - 1), out=b, dtype=b.dtype)
            return b
        a = np.array(a, _dtype2(kind, m, a.dtype.itemsize), copy=False)
        a *= (2**m - 1) // (2**n - 1)
        return a
    # upscale to a multiple of n bits,
    # then downscale with precision loss
    prec_loss(dtypeobj_in, dtypeobj)
    o = (m // n + 1) * n
    if copy:
        b = np.empty(a.shape, _dtype2(kind, o))
        np.multiply(a, (2**o - 1) // (2**n - 1), out=b, dtype=b.dtype)
        b //= 2 ** (o - m)
        return b
    a = np.array(a, _dtype2(kind, o, a.dtype.itemsize), copy=False)
    a *= (2**o - 1) // (2**n - 1)
    a //= 2 ** (o - m)
    return a



# -*- coding: utf-8 -*-
"""Utility functions that test whether things are true or not."""

__all__ = [
    "all_size",
    "all_type",
    "isanynone",
    "isComparable",
    "isiterable",
    "isLikeList",
    "isnone",
    "isproperty",
    "isTuple",
]
from typing import Optional, Iterable as IterableType, Tuple, Union, Any

from collections.abc import Iterable
from numpy import ndarray, dtype, isnan, logical_and  # pylint: disable=redefined-builtin

from ..compat import string_types
from ..core.Typing import NumericArray


def all_size(iterator: IterableType, size: Optional[Union[int, Tuple]] = None) -> bool:
    """Check whether each element of *iterator* is the same length/shape.

    Arguments:
        iterator (Iterable): list or other iterable of things with a length or shape

    Keyword Arguments:
        size(int, tuple or None): Required size of each item in iterator.

    Returns:
        True if all objects are the size specified (or the same size if size is None).
    """
    if hasattr(iterator[0], "shape"):
        sizer = lambda x: x.shape
    else:
        sizer = len

    if size is None:
        size = sizer(iterator[0])
    ret = False
    for i in iterator:
        if sizer(i) != size:
            break
    else:
        ret = True
    return ret


def all_type(iterator: IterableType, typ: type) -> bool:
    """Determine if an interable omnly contains a common type.

    Arguments:
        iterator (Iterable):
            The object to check if it is all iterable
        typ (class):
            The type to check for.

    Returns:
        True if all elements are of the type typ, or False if not.

    Notes:
        Routine will iterate the *iterator* and break when an element is not of
        the search type *typ*.
    """
    ret = False
    if isinstance(iterator, ndarray):  # Try to short circuit for arrays
        try:
            return iterator.dtype == dtype(typ)
        except TypeError:
            pass
    if isiterable(iterator):
        for i in iterator:
            if not isinstance(i, typ):
                break
        else:
            ret = True
    return ret


def isanynone(*args: Any) -> bool:
    """Intelligently check whether any of the inputs are None."""
    for arg in args:
        if arg is None:
            return True
    return False


def isComparable(v1: NumericArray, v2: NumericArray) -> bool:
    """Return true if v1 and v2 can be compared sensibly.

    Args:
        v1,v2 (any):
            Two values to compare

    Returns:
        False if both v1 and v2 are numerical and both nan, otherwise True.
    """
    try:
        return not (isnan(v1) and isnan(v2))
    except TypeError:
        return True
    except ValueError:
        try:
            return not logical_and(isnan(v1), isnan(v2)).any()
        except TypeError:
            return False


def isiterable(value: Any) -> bool:
    """Chack to see if a value is iterable.

    Args:
        value :
            Entity to check if it is iterable

    Returns:
        (bool):
            True if value is an instance of collections.Iterable.
    """
    return isinstance(value, Iterable)


def isLikeList(value: Any) -> bool:
    """Return True if value is an iterable but not a string."""
    return isiterable(value) and not isinstance(value, string_types)


def isnone(iterator: Optional[IterableType]) -> bool:
    """Return True if input is None or an empty iterator, or an iterator of None.

    Args:
        iterator (None or Iterable):

    Returns:
        True if iterator is None, empty or full of None.
    """
    if iterator is None:
        ret = True
    elif isiterable(iterator) and not isinstance(iterator, string_types):
        try:
            l = len(iterator)
        except TypeError:
            l = 0
        if l == 0:  # pylint: disable=len-as-condition
            ret = True
        else:
            for i in iterator:
                if i is not None:
                    ret = False
                    break
            else:
                ret = True
    else:
        ret = False
    return ret


def isproperty(obj: Any, name: str) -> bool:
    """Check whether an attribute of an object or class is a property.

    Args:
        obj (instance or class):
            Thing that has the attribute to check
        name (str):
            Name of the attribute that might be a property

    Returns:
        (bool):
            Whether the name is a property or not.
    """
    if not isinstance(obj, type):
        obj = type(obj)
    elif not issubclass(obj, object):
        raise TypeError(f"Can only check for property status on attributes of an object or a class not a {type(obj)}")
    return hasattr(obj, name) and isinstance(getattr(obj, name), property)


def isTuple(obj: Any, *args: type, strict: bool = True) -> bool:
    """Determine if obj is a tuple of a certain signature.

    Args:
        obj(object):
            The object to check
        *args(type):
            Each of the succeeding arguments are used to determine the expected type of each element.

    Keywoprd Arguments:
        strict(bool):
            Whether the elements of the tuple have to be exactly the type specified or just castable as the type

    Returns:
        (bool):
            True if obj is a matching tuple.
    """
    if not isinstance(obj, tuple):
        return False
    if args and len(obj) != len(args):
        return False
    for t, e in zip(args, obj):
        if strict:
            if not isinstance(e, t):
                bad = True
                break
        else:
            try:
                _ = t(e)
            except ValueError:
                bad = True
                break
    else:
        bad = False
    return not bad



"""Simple use of lmfit to fit data."""

# pylint: disable=invalid-name
from numpy import linspace, random

from Stoner import Data
from Stoner.analysis.fitting.models.generic import PowerLaw
from Stoner.analysis.fitting import odr_Model

# Make some data
x = linspace(0, 10.0, 101)
y = PowerLaw().func(x, 1e-1, 1.64) * random.normal(
    loc=1.0, scale=0.05, size=101
)
d = Data(x, y, column_headers=["Time", "Signal"], setas="xy")

# Do the fitting and plot the result
fit = d.curve_fit(
    PowerLaw,
    result=True,
    header="LM-Model Fit",
    residuals=True,
    output="report",
)

ODRModel = odr_Model(PowerLaw, p0=(1, 1))
fit = d.curve_fit(
    ODRModel,
    result=True,
    header="ODR-Fit",
    residuals=True,
    output="report",
    prefix="ODRModel",
)
# Reset labels
d.labels = []
# Make nice two panel plot layout
d.figure(figsize=(7, 5), no_axes=True)
ax = d.subplot2grid((3, 1), (2, 0))
d.setas = "x..y"
d.plot(fmt="g+", label="Fit residuals")
d.setas = "x....y"
d.plot(fmt="b+", label="ODRModel Residuals")
d.title = ""

ax = d.subplot2grid((3, 1), (0, 0), rowspan=2)
d.setas = "xyy.y"
d.plot(fmt=["ro", "g-", "b-"])
d.xticklabels = [[]]
d.ax_xlabel = ""

# Annotate plot with fitting parameters
d.annotate_fit(PowerLaw, x=0.1, y=0.25, fontdict={"size": "x-small"})
d.annotate_fit(
    ODRModel, x=0.65, y=0.15, fontdict={"size": "x-small"}, prefix="ODRModel"
)
d.title = "curve_fit with models"



"""Fit a sphere with curve_fit."""

# pylint: disable=invalid-name, redefined-outer-name
from numpy import (
    sin,
    cos,
    pi,
    column_stack,
    zeros_like,
    ones_like,
    meshgrid,
    linspace,
)
from numpy.random import normal, uniform, seed
import matplotlib.pyplot as plt

from Stoner import Data

seed(12345)  # Ensure consistent random numbers!


def transform(r, q, p):
    """Converts from spherical to cartesian coordinates."""
    x = r * cos(q) * cos(p)
    y = r * cos(q) * sin(p)
    z = r * sin(q)
    return x, y, z


def sphere(coords, a, b, c, r):
    """Returns zero if (x,y,z) lies on a sphere centred at (a,b,c) with radius r."""
    x, y, z = coords.T
    return (x - a) ** 2 + (y - b) ** 2 + (z - c) ** 2 - r**2


# Create some points approximately spherical distribution
num = 25
q = uniform(low=-pi / 2, high=pi / 2, size=num)
p = uniform(low=0, high=2 * pi, size=num)
r = normal(loc=3.0, size=num, scale=0.1)

x, y, z = transform(r, q, p)

x += 3.0
y -= 4.0
z += 2.0

# Construct the  DataFile object
d = Data(
    column_stack((x, y, z)),
    setas="xyz",
    filename="Best fit sphere",
    column_headers=["X", "Y", "Z"],
)
d.template.fig_width = 5.2
d.template.fig_height = 5.0  # Square aspect ratio
d.plot_xyz(plotter="scatter", marker=",", griddata=False)
d.set_box_aspect((1, 1, 1.0))  # Passing through to the current axes
# curve_fit does the hard work
popt, pcov = d.curve_fit(sphere, (0, 1, 2), zeros_like(d.x))

# This manually constructs the best fit sphere
a, b, c, r = popt
p = linspace(-pi / 2, pi / 2, 16)
q = linspace(-pi, pi, 31)
P, Q = meshgrid(p, q)
R = ones_like(P) * r
x, y, z = transform(R, Q, P)
x += a
y += b
z += c

ax = d.axes[0]
ax.plot_surface(
    x, y, z, rstride=1, cstride=1, color=(1.0, 0.0, 0.0, 0.25), linewidth=0
)
plt.draw()



"""Demonstrate STXM Image Processing - G.Burnell Nov. 2017"""

# pylint: disable=invalid-name,no-member
from os.path import join, dirname
from types import MethodType

import numpy as np
import matplotlib.pyplot as plt
from lmfit.models import LorentzianModel

from Stoner.Image import ImageFolder
from Stoner.HDF5 import STXMImage

# Load the images
thisdir = dirname(__file__)

imgs = ImageFolder(type=STXMImage)
for fname in [
    "Sample_Image_2017-10-15_100.hdf5",
    "Sample_Image_2017-10-15_101.hdf5",
]:
    # Load the image
    img = STXMImage(join(thisdir, "..", "..", "..", "sample-data", fname))
    a1 = id(img.metadata)
    img.gridimage()
    a2 = id(img.metadata)
    img.crop(5, -15, 5, -5, _=True)  # regularise grid and crop
    a3 = id(img.metadata)
    imgs += img
    a4 = [id(i.metadata) for i in imgs]

# Align the two images
imgs.align(method="imreg_dft", scale=10)
imgs.each.crop(2)

# Calculate the XMCD image
xmcd = imgs[0] // imgs[1]
xmcd.normalise()

imgs += xmcd
strctural = (imgs[0] + imgs[1]) / 2
imgs += strctural

# Mask out the background, using the structural image
xmcd.mask = strctural.image > strctural.threshold_otsu()

# Remove the background XMCD atio
xmcd.mask.invert()
mean = xmcd.mean()
xmcd.mask = False
xmcd -= mean
xmcd.mask = strctural.image > strctural.threshold_otsu()
xmcd.normalise()

# Create a profile and plot it
profile = xmcd.profile_line((0, 0), (100, 100))
profile.figure(figsize=(7, 6), no_axes=True)
profile.subplot(222)
profile.plot()
profile.title = "XMCD Cross Section"

# Show the images as well
profile.subplot(221)
strctural.imshow(figure=profile.fig, title="Structural Image")

profile.subplot(223)
xmcd.imshow(figure=profile.fig, title="XMCD Image")

# Make a histogram of the intensity values
hist = xmcd.hist(bins=200)
hist.column_headers = ["XMCD Signal", "Frequency"]
hist.labels = None
hist.fig = profile.fig

# Construct a two Lorentzian peak model
peak1 = LorentzianModel(prefix="peak1_")
peak2 = LorentzianModel(prefix="peak2_")
params = peak1.make_params()
params.update(peak2.make_params())
double_peak = peak1 + peak2


def guess(self, data, **kwargs):
    """Function to guess the parameters of two Lorentzian peaks."""
    x = kwargs.get("x")
    l = len(data)
    i1 = np.argmax(data[: l // 2])  # Location of first peak
    i2 = np.argmax(data[l // 2 :]) + l // 2

    params = self.make_params()
    for k, p in zip(
        params,
        [
            data[i1] / np.sqrt(2),
            x[i1],
            0.25,
            0.5,
            data[i1],
            data[i2] / np.sqrt(2),
            x[i2],
            0.25,
            0.5,
            data[i2],
        ],
    ):
        params[k].value = p
    return params


# This shows how to replace the default guess (which is useless) with our function guess
double_peak.guess = MethodType(guess, double_peak)

# Set the initial values with sensible starting points guessed from the data
# Fit the intensity values and add to thedata file
res = hist.lmfit(double_peak, output="report")
hist.add_column(res.init_fit, header="Initial Fit")
hist.add_column(res.best_fit, header="Best Fit")

# Plot the results as well
hist.setas = "xyyy"
profile.subplot(224)
hist.plot(fmt=["b+", "b--", "r-"])
hist.title = "Intensity histogram"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""":py:class:`lmfit.Model` model classes and functions for various models of electron tunnelling."""
# pylint: disable=invalid-name
__all__ = [
    "BDR",
    "FowlerNordheim",
    "Model",
    "Simmons",
    "TersoffHammann",
    "bdr",
    "fowlerNordheim",
    "simmons",
    "tersoffHammann",
]

import numpy as np

from lmfit import Model
from lmfit.models import update_param_vals


def simmons(V, A, phi, d):
    """Simmons model of electron tunnelling.

    Args:
        V (array): Bias voltage
        A (float): Area of barrier in micron^2
        phi (float): barrier height in eV
        d (float): barrier width in angstroms

    Return:
        Data for tunneling rate according to the Sommons model.

    .. note::

        Simmons model from Simmons J. App. Phys. 34 6 1963

    Example:
        .. plot:: samples/Fitting/Simmons.py
            :include-source:
            :outname: simmons
    """
    I = (
        6.2e2
        * A
        / d**2
        * (
            (phi - V / 2) * np.exp(-1.025 * d * np.sqrt(phi - V / 2))
            - (phi + V / 2) * np.exp(-1.025 * d * np.sqrt(phi + V / 2))
        )
    )
    return I


def bdr(V, A, phi, dphi, d, mass):
    """BDR model tunnelling.

    Args:
        V (array): ias voltage
        A (float): barrier area in m^2
        phi (float): average barrier height in eV
        dphi (float): change in barrier height in eV
        d (float): barrier width in angstrom
        mass (float): effective electron mass as a fraction of electron rest mass

    Return:
        Data for tunneling rate  according to the BDR model.
    .. note::

       See Brinkman et. al. J. Appl. Phys. 41 1915 (1970) or Tuan Comm. in Phys. 16, 1, (2006)

    Example:
        .. plot:: samples/Fitting/BDR.py
            :include-source:
            :outname: bdr
    """
    mass = abs(mass)
    phi = abs(phi)
    d = abs(d)
    I = (
        3.16e10
        * A**2
        * np.sqrt(phi)
        / d
        * np.exp(-1.028 * np.sqrt(phi) * d)
        * (V - 0.0214 * np.sqrt(mass) * d * dphi / phi**1.5 * V**2 + 0.0110 * mass * d**2 / phi * V**3)
    )
    return I


def fowlerNordheim(V, A, phi, d):
    """Fowler Nordhiem Model of electron tunnelling.

    Args:
        V (array): Bias voltage
        A (float): Area of barrier in m^2
        phi (float): barrier height in eV
        d (float): barrier width in angstroms

    Return:
        Tunneling rate according to Fowler Nordheim model.

    Example:
        .. plot:: samples/Fitting/FowlerNordheim.py
            :include-source:
            :outname: fowlernordheim
    """
    I = V / np.abs(V) * 3.38e6 * A * V**2 / (d**2 * phi) * np.exp(-0.689 * phi**1.5 * d / np.abs(V))
    return I


def tersoffHammann(V, A):
    """Tersoff-Hamman model for tunnelling through STM tip.

    Args:
        V (array): bias voltage
        A (float): Tip conductance

    Return:
        A linear fit.
    """
    I = A * V
    return I


class Simmons(Model):
    """Simmons model of electron tunnelling.

    Args:
        V (array): Bias voltage
        A (float): Area of barrier in m^2
        phi (float): barrier height in eV
        d (float): barrier width in angstroms

    Return:
        Data for tunneling rate according to the Sommons model.

    .. note::

       Simmons model from Simmons J. App. Phys. 34 6 1963

    Example:
        .. plot:: samples/Fitting/Simmons.py
            :include-source:
            :outname: simmons-class
    """

    display_names = ["A", r"\phi", "d"]

    def __init__(self, *args, **kwargs):
        """Configure Initial fitting function."""
        super().__init__(simmons, *args, **kwargs)

    def guess(self, data, V=None, **kwargs):  # pylint: disable=unused-argument
        """Set the A, phi and d values to typical answers for a small tunnel junction."""
        pars = self.make_params(A=1e3, phi=3.0, d=10.0)
        return update_param_vals(pars, self.prefix, **kwargs)


class BDR(Model):
    """BDR model tunnelling.

    Args:
        V (array): ias voltage
        A (float): barrier area in m^2
        phi (float): average barrier height in eV
        dphi (float): change in barrier height in eV
        d (float): barrier width in angstrom
        mass (float): effective electron mass as a fraction of electron rest mass

    Return:
        Data for tunneling rate  according to the BDR model.

    .. note::

       See Brinkman et. al. J. Appl. Phys. 41 1915 (1970) or Tuan Comm. in Phys. 16, 1, (2006)

    Example:
        .. plot:: samples/Fitting/BDR.py
            :include-source:
            :outname: bdr-class
    """

    def __init__(self, *args, **kwargs):
        """Configure Initial fitting function."""
        super().__init__(bdr, *args, **kwargs)

    def guess(self, data, V=None, **kwargs):  # pylint: disable=unused-argument
        """Set the A, phi,dphi,d and mass values to typical answers for a small tunnel junction."""
        pars = self.make_params(A=1e-12, phi=3.0, d=10.0, dphi=1.0, mass=1.0)
        return update_param_vals(pars, self.prefix, **kwargs)


class FowlerNordheim(Model):
    """Fowler Nordhiem Model of electron tunnelling.

    Args:
        V (array): Bias voltage
        A (float): Area of barrier in m^2
        phi (float): barrier height in eV
        d (float): barrier width in angstroms

    Return:
        Tunneling rate according to Fowler Nordheim model.

    Example:
        .. plot:: samples/Fitting/FowlerNordheim.py
            :include-source:
            :outname: fowlernordheim-class
    """

    def __init__(self, *args, **kwargs):
        """Configure Initial fitting function."""
        super().__init__(fowlerNordheim, *args, **kwargs)

    def guess(self, data, V=None, **kwargs):  # pylint: disable=unused-argument
        """Set the A, phi and d values to typical answers for a small tunnel junction."""
        pars = self.make_params(A=1e-12, phi=3.0, d=10.0)
        return update_param_vals(pars, self.prefix, **kwargs)


class TersoffHammann(Model):
    """Tersoff-Hamman model for tunnelling through STM tip.

    Args:
        V (array): bias voltage
        A (float): Tip conductance

    Return:
        A linear fit.
    """

    def __init__(self, *args, **kwargs):
        """Configure Initial fitting function."""
        super().__init__(tersoffHammann, *args, **kwargs)

    def guess(self, data, V=None, **kwargs):
        """Set the parameter values from an apporximate line."""
        pars = self.make_params(A=np.mean(data / V))
        return update_param_vals(pars, self.prefix, **kwargs)



# -*- coding: utf-8 -*-
"""Sample script for fitting BlochGrueneisen function to a data file.

@author: phygbu
"""
# pylint: disable=invalid-name
import re

import numpy as np

from Stoner import Data
from Stoner.analysis.fitting.models.e_transport import BlochGrueneisen


def select_col(data, message):
    """Select the column with the data."""
    print("Unable to guess column")
    for i, col in enumerate(data.column_headers):
        print("{} : {}".format(i, col))
    while True:
        try:
            return int(input(message))
        except ValueError:
            print("Please select a column number")
            continue


# Load a datafile
d = Data(False)

# These regular expressions are intended to match a tempeerature and resistance coluimns
t_pat = [re.compile(r"^[Tt][\s\(]"), re.compile(r"[Tt]emp")]
r_pat = [re.compile(r"[Rr]ho"), re.compile(r"[Rr]es")]

# Search for each column
for pat in t_pat:
    t_col = d.find_col(pat, force_list=True)
    if len(t_col) == 1:
        t_col = t_col[0]
        break
else:
    t_col = select_col(d, "Select column for temperature data :")

for pat in r_pat:
    r_col = d.find_col(pat, force_list=True)
    if len(r_col) == 1:
        r_col = r_col[0]
        break
else:
    r_col = select_col(d, "Select column for resistance data :")

# Configure the columns in the setas attribute
d.mask = np.isnan(d.data)
d.del_rows()
d.setas(x=t_col, y=r_col)
# Initialise the model, set the n parmeter to be fixed and show our guesses
model = BlochGrueneisen()
model.param_hints["n"] = {"vary": False, "value": 5}
print("Initial guesses: {}".format(model.guess(d.y, x=d.x)))

# Do the fit
popt, pcov = d.lmfit(model, absolute_sigma=False, result=True, header="Bloch")

# Add the Bloch column to the setas
d.setas(Bloch="y", reset=False)
# Make a plot
d.plot(
    fmt=["r.", "b-"], label=["Data", r"$Bloch-Gr\"ueisen Fit$"], markersize=1
)
d.xlabel = "Temperature (K)"
d.ylabel = "Resistance ($\Omega$)"
d.annotate_fit(model, x=0.05, y=0.35, mode="eng")



"""Demo of new Stoner.Analysis.AnalyseFile.lmfit."""

# pylint: disable=invalid-name
from os.path import join

from Stoner.analysis.fitting.models import cfg_data_from_ini, cfg_model_from_ini
from Stoner import __home__

config = join(__home__, "..", "scripts", "PCAR-chi^2.ini")
datafile = join(__home__, "..", "sample-data", "PCAR Co Data.csv")

d = cfg_data_from_ini(config, datafile)
model, p0 = cfg_model_from_ini(config, data=d)

fit = d.lmfit(model, p0=p0, result=True, header="Fit", output="data")
fit.plot(multiple="panels", capsize=3)
fit.yscale = "log"  # Adjust y scale for chi^2



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""":py:class:`lmfit.Model` model classes and functions for various generic models."""
# pylint: disable=invalid-name
__all__ = [
    "Linear",
    "Lorentzian_diff",
    "Model",
    "PowerLaw",
    "Quadratic",
    "StretchedExp",
    "linear",
    "lorentzian_diff",
    "powerLaw",
    "quadratic",
    "stretchedExp",
]

import numpy as np

from lmfit import Model
from lmfit.models import LinearModel as _Linear  # NOQA pylint: disable=unused-import
from lmfit.models import PowerLawModel as _PowerLaw  # NOQA pylint: disable=unused-import
from lmfit.models import QuadraticModel as _Quadratic  # NOQA pylint: disable=unused-import
from lmfit.models import update_param_vals


def linear(x, intercept, slope):
    """Calculate a linear function."""
    return slope * x + intercept


def quadratic(x, a, b, c):
    r"""Calculate a simple quadratic fitting function.

    Args:
        x (array): Input data
        a (float): Quadratic term co-efficient
        b (float): Linear term co-efficient
        c (float): Constant offset term

    Returns:
        Array of data.

    :math:`y=ax^2+bx+c`

    Example:
        .. plot:: samples/Fitting/Quadratic.py
            :include-source:
            :outname: quadratic
    """
    return a * x**2 + b * x + c


def powerLaw(x, A, k):
    r"""Power Law Fitting Equation.

    Args:
        x (array): Input data
        A (float): Prefactor
        k (float): Power

    Return:
        Power law.

    :math:`p=Ax^k`

    Example:
        .. plot:: samples/Fitting/Powerlaw.py
            :include-source:
            :outname: powerlaw
    """
    return A * x**k


def stretchedExp(x, A, beta, x_0):
    r"""Calculate a stretched exponential fuinction.

    Args:
        x (array): x data values
        A (float): Constant prefactor
        beta (float): Stretch factor
        x_0 (float): Scaling factor for x data

    Return:
        Data for a stretched exponentional function.

    The stretched exponential is defined as :math:`y=A\exp\left[\left(\frac{-x}{x_0}\right)^\beta\right]`.
    """
    return A * np.exp(-((x / x_0) ** beta))


def lorentzian_diff(x, A, sigma, mu):
    r"""Implement a differential form of a Lorentzian peak.

    Args:
        x (array): x data
        A (float): Peak amplitude
        sigma (float): peak wideth
        mu (float): peak location in x

    Returns
        :math:`\frac{A \sigma \left(2 \mu - 2 x\right)}{\pi \left(\sigma^{2} +
                                                                      \left(- \mu + x\right)^{2}\right)^{2}}`

    Example:
        .. plot:: samples/Fitting/lorentzian.py
            :include-source:
            :outname: lorentzian_diff_func
    """
    return A * sigma * (2 * mu - 2 * x) / (np.pi * (sigma**2 + (-mu + x) ** 2) ** 2)


class Linear(_Linear):
    """Simple linear fit class."""


class Quadratic(_Quadratic):
    r"""A Simple quadratic fitting function.

    Args:
        x (array): Input data
        a (float): Quadratic term co-efficient
        b (float): Linear term co-efficient
        c (float): Constant offset term

    Returns:
        Array of data.

    :math:`y=ax^2+bx+c`

    Example:
        .. plot:: samples/Fitting/Quadratic.py
            :include-source:
            :outname: quadratic-class
    """


class PowerLaw(_PowerLaw):
    r"""Power Law Fitting Equation.

    Args:
        x (array): Input data
        A (float): Prefactor
        k (float): Power

    Return:
        Power law.

    :math:`p=Ax^k`

    Example:
        .. plot:: samples/Fitting/Powerlaw.py
            :include-source:
            :outname: powerlaw-class
    """


class StretchedExp(Model):
    r"""A stretched exponential fuinction.

    Args:
        x (array): x data values
        A (float): Constant prefactor
        beta (float): Stretch factor
        x_0 (float): Scaling factor for x data

    Return:
        Data for a stretched exponentional function.

    The stretched exponential is defined as :math:`y=A\exp\left[\left(\frac{-x}{x_0}\right)^\beta\right]`.
    """

    display_names = ["A", r"\beta", "x_0"]

    def __init__(self, *args, **kwargs):
        """Configure Initial fitting function."""
        super().__init__(stretchedExp, *args, **kwargs)

    def guess(self, data, x=None, **kwargs):
        """Guess parameters for the stretched exponential from data."""
        A, beta, x0 = 1.0, 1.0, 1.0
        if x is not None:
            A = data[np.argmin(np.abs(x))]
            x = np.log(x)
            y = np.log(-np.log(data / A))
            d = np.column_stack((x, y))
            d = d[~np.isnan(d).any(axis=1)]
            d = d[~np.isinf(d).any(axis=1)]
            d1, d2 = np.polyfit(d[:, 0], d[:, 1], 1)
            beta = d1
            x0 = np.exp(d2 / beta)
        pars = self.make_params(A=A, beta=beta, x_0=x0)
        return update_param_vals(pars, self.prefix, **kwargs)


class Lorentzian_diff(Model):
    r"""Provides a lmfit Model rerprenting the differential form of a Lorentzian Peak.

    Args:
        x (array): x data
        A (float): Peak amplitude
        sigma (float): peak wideth
        mu (float): peak location in x

    Returns
        :math:`\frac{A \sigma \left(2 \mu - 2 x\right)}{\pi \left(\sigma^{2} +
                                                                      \left(- \mu + x\right)^{2}\right)^{2}}`

    Example:
        .. plot:: samples/Fitting/lorentzian.py
            :include-source:
            :outname: lorentzian_diff_class
    """

    display_names = ["A", r"\sigma", r"\mu"]

    def __init__(self, *args, **kwargs):
        """Configure Initial fitting function."""
        super().__init__(lorentzian_diff, *args, **kwargs)

    def guess(self, data, x=None, **kwargs):
        """Guess parameters as gamma=2, H_k=0, M_s~(pi.f)^2/(mu_0^2.H)-H."""
        if x is None:
            x = np.linspace(1, len(data), len(data) + 1)

        x1 = x[np.argmax(data)]
        x2 = x[np.argmin(data)]
        sigma = abs(x1 - x2)
        mu = (x1 + x2) / 2.0
        y1 = np.max(data)
        y2 = np.min(data)
        dy = y1 - y2
        A = dy * (4 * np.pi * sigma**2) / (3 * np.sqrt(3))

        pars = self.make_params(A=A, sigma=sigma, mu=mu)
        pars["A"].min = 0
        pars["sigma"].min = 0
        pars["mu"].min = np.min(x)
        pars["mu"].max = np.max(x)
        return update_param_vals(pars, self.prefix, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 20 21:22:18 2018

@author: phygbu
"""
from Stoner.Image import ImageFile, ImageFolder, ImageStack
import numpy as np
import os
import Stoner
import pytest

Stoner.Options.multiprocessing = False

testdir = os.path.join(os.path.dirname(__file__), "coretestdata", "testims")

istack2 = ImageStack()
for theta in np.linspace(0, 360, 16):
    i = ImageFile(np.zeros((100, 100)))
    x, y = 10 * np.cos(np.pi * theta / 180) + 50, 10 * np.sin(np.pi * theta / 180) + 50
    i.draw.circle(x, y, 25)
    i.filename = "Angle {}".format(theta)
    istack2.insert(0, i)


selftd = ImageFolder(testdir, pattern="*.png")
selfks = ImageStack(testdir)
selfks = ImageStack(selftd)  # load in two ways
assert len(selfks) == len(os.listdir(testdir))
selfistack2 = istack2.clone


def test_ImageStack_align():
    assert selfistack2.shape == (16, 100, 100), "ImageStack.shape wrong at {}".format(selfistack2.shape)
    i = ImageFile(np.zeros((100, 100))).draw.circle(50, 50, 25)

    istack2 = selfistack2.clone
    istack2.align(i, method="chi2_shift")
    data = istack2.metadata.slice(["tvec"], output="Data")
    assert data.shape == (16, 2), "Slice metadata went a bit funny"

    istack2 = selfistack2.clone
    with pytest.raises(ValueError):
        istack2.align(0, 10, 20, method="imreg_dft")
    with pytest.raises(TypeError):
        istack2.align(np.zeros(5), method="imreg_dft")
    with pytest.raises(TypeError):
        istack2.align(object, method="imreg_dft")

    istack2.align(method="imreg_dft")
    assert istack2[0]["tvec"] == (0.0, 0.0), "stack didn't align to first image"

    istack2 = selfistack2.clone

    istack2.align(4)
    data = istack2.metadata.slice(["tvec", "angle", "scale"], output="Data")
    assert data.shape == (16, 4), "Slice metadata went a bit funny"
    assert sorted(data.column_headers) == [
        "angle",
        "scale",
        "tvec[0]",
        "tvec[1]",
    ], "slice metadata column headers wrong at {}".format(data.column_headers)


@pytest.mark.filterwarnings("ignore:.*:UserWarning")
def test_ImageStack_methods():
    istack2 = selfistack2.clone
    m1 = selfistack2.mean().crop(10)
    istack2.align(i, method="imreg_dft")
    data = istack2.metadata.slice(["tvec", "angle", "scale"], output="Data")
    assert data.shape == (16, 4), "Slice metadata went a bit funny"
    assert sorted(data.column_headers) == [
        "angle",
        "scale",
        "tvec[0]",
        "tvec[1]",
    ], "slice metadata column headers wrong at {}".format(data.column_headers)

    istack2.each.crop("translation_limits")
    m2 = istack2.mean()
    std = istack2.stddev()
    err = istack2.stderr()

    assert std.mean() < 0.01, "To big a standard deviation in stack after align"
    assert err.mean() < 0.003, "To big a standard error in stack after align"
    assert istack2.shape == (16, 80, 80), "Stack translation_limits and crop failed."

    assert np.abs(m1.mean() - m2.mean()) / m1.mean() < 1e-2, "Problem calculating means of stacks."
    s1 = selfistack2[:, 45:55, 45:55]
    s2 = selfistack2[:, 50, :]
    s3 = selfistack2[:, :, 50]
    s4 = selfistack2[5, :, :]
    assert s1.shape == (16, 10, 10), "3D slicing to produce 3D stack didn't work."
    assert s2.shape == (16, 100), "3D slicing to 2D section z-y plane failed."
    assert s3.shape == (16, 100), "3D slicing to 2D section z-x plane failed."
    assert s4.shape == (100, 100), "3D slicing to 2D section x-y plane failed."
    assert len(selfistack2) == 16, "len(ImageFolder.images) failed."
    sa = []
    for im in selfistack2.images:
        sa.append(im.shape)
    sa = np.array(sa)
    assert np.all(sa == np.ones((16, 2)) * 100), "Result from iterating over images failed."
    selfistack2.each.normalise()
    assert (np.array(selfistack2.each.min()).mean(), np.array(selfistack2.each.max()).mean()) == (
        -1.0,
        1.0,
    ), "Adjust contrast failure"
    im1 = selfistack2[0]
    im1.normalise()
    im1.convert(np.int32)
    im2 = im1.convert(np.float32, force_copy=True)
    conv_err = (selfistack2[0].image - im2.image).max()
    assert conv_err < 1e-7, "Problems double converting images:{}.".format(conv_err)
    im1 = selfistack2[0].convert(np.int64)
    im1 = im1.convert(np.int8)
    im2 = selfistack2[0].convert(np.int8)
    assert abs((im2 - im1).max()) <= 2.0, "Failed up/down conversion to integer images."


def test_init():
    # try to init with a few different call sequences
    listinit = []
    for i in range(10):
        listinit.append(np.arange(12).reshape(3, 4))
    npinit = np.arange(1000).reshape(5, 10, 20)
    listinit = ImageStack(listinit)
    assert listinit.shape == (10, 3, 4), "problem with initialising ImageStack with list of data"
    npinitist = ImageStack(npinit)
    assert np.allclose(npinitist.imarray, npinit), "problem initiating with 3d numpy array"
    ist2init = ImageStack(selfistack2)
    assert np.allclose(ist2init.imarray, selfistack2.imarray), "problem initiating with other ImageStack"
    assert all(
        [k in ist2init[0].metadata.keys() for k in selfistack2[0].metadata.keys()]
    ), "problem with metadata when initiating with other ImageStack"
    imfinit = ImageStack(selftd)  # init with another ImageFolder
    assert len(imfinit) == 8, "Couldn't load from another ImageFolder object"


def test_accessing():
    # ensure we can write and read to the stack in different ways
    ist2 = ImageStack(np.arange(60).reshape(4, 3, 5))
    im = np.zeros((3, 5), dtype=int)
    ist2[0].image = im
    ist2[1] = im
    ist2[2].image[:] = im
    ist2[3][0, 0] = 100
    ist2[3].image[0, 1] = 101
    # ist2[3,0,2] = 102 may want to support this type of index accessing in the future?
    for i in range(3):
        assert np.allclose(ist2[i].asarray(), im)
    assert np.allclose(ist2[3][0, 0], 100)
    assert np.allclose(ist2[3][0, 1], 101)
    # check imarray behaviour
    ist2.imarray = np.arange(60).reshape(4, 3, 5) * 3
    assert np.allclose(ist2.imarray, np.arange(60).reshape(4, 3, 5) * 3)
    ist2.imarray[1, 2, 3] = 500
    assert ist2.imarray[1, 2, 3] == 500
    # check slice access #not implemented yet
    # im2 = np.zeros((2,3,5))
    # ist2[3:5] = im2 #try setting two images at once
    # assert np.allclose(ist2[4],im))


def test_methods():
    # check function generator machinery works
    selfistack2.each.crop(0, 30, 0, 50)
    assert selfistack2.shape == (16, 50, 30), "Unexpected size of imagestack got {} for 16x50x30".format(
        selfistack2.shape
    )
    ist2 = ImageStack(np.arange(60).reshape(4, 3, 5))
    assert issubclass(
        ist2.imarray.dtype.type, np.integer
    ), "Unexpected dtype in image stack2 got {} not int32".format(ist2.imarray.dtype)
    t1 = ImageStack(np.arange(60).reshape(4, 3, 5))
    t1.asfloat(normalise=False, clip_negative=False)
    assert t1.imarray.dtype == np.float64, f"Type Failure {t1.imarray.dtype}"
    assert np.max(t1.imarray) == 59.0
    t2 = ImageStack(np.arange(60).reshape(4, 3, 5))
    t2.asfloat(normalise=True, clip_negative=True)
    # assert  np.max(t2.imarray) == (2*59+1)/(2**31-(-2**31)) )
    assert np.min(t2.imarray) >= 0
    ist3 = ist2.clone
    assert not np.may_share_memory(ist2.imarray, ist3.imarray)
    del ist3[-1]
    assert len(ist3) == len(ist2) - 1
    assert np.allclose(ist3[0], ist2[0])
    ist3.insert(1, np.arange(18).reshape(3, 6))
    assert ist3[1].shape == (3, 6), "inserting an image of different size to stack"
    im1 = ImageFile(np.zeros((100, 100)))
    im1.draw.circle(50, 50, 25)
    im2 = im1.clone
    pi = np.pi
    X, Y = np.mgrid[-pi : pi : pi / 50, -pi : pi : pi / 50]
    bground = ImageFile(np.sin(X) + np.cos(Y))
    im2.image += bground.image
    bground.mask = im1.image == 1.0
    ist = ImageStack() + im2
    ist.subtract(bground)
    assert np.all(np.isclose(ist[0].image, im1.image))


def test_clone():
    ist2 = ImageStack(np.arange(60).reshape(4, 3, 5))
    ist3 = ist2.clone
    assert not np.may_share_memory(ist2.imarray, ist3.imarray)  # Check new memory has been allocated for clone
    del ist3[-1]
    assert len(ist3) == len(ist2) - 1  # Check deletion operation only happens for clone
    assert np.allclose(ist3[0], ist2[0])  # Check values are the same
    ist3.insert(1, np.arange(18).reshape(3, 6))  # Insert a larger image
    assert ist3[1].shape == (3, 6), "inserting an image of different size to stack"


def test_operators():
    im1 = ImageFile(np.zeros((100, 100)) + 0.75)
    im2 = ImageFile(np.zeros((100, 100)) + 0.25)
    ist1 = ImageStack() + im1
    ist2 = ImageStack() + im2
    ist3 = ist1 // ist2
    assert np.all(np.isclose(ist3[0].image, np.ones((100, 100)) * 0.5))
    im1a = ImageFile(np.ones((100, 100), dtype=int))
    with pytest.raises(ValueError):
        ist1a = ImageStack() + im1a
        _ = ist1a // ist2
    im1b = im1.clone
    im1b.image = im1b.image[1:-1, 1:-1]
    ist1b = ImageStack() + im1b
    with pytest.raises(ValueError):
        _ = ist1b // ist2


def test_mask():
    im = ImageFile(np.arange(12).reshape(3, 4))
    im.mask = np.zeros(im.shape, dtype=bool)
    im.mask.data[1, 1] = True
    ist2 = ImageStack(np.arange(60).reshape(4, 3, 5))
    ist2.insert(1, im)  # Insert an image with a mask
    assert ist2[1].mask.data.shape == ist2[1].shape
    assert ist2[1].mask[1, 1], "inserting an image with a mask into an ImageStack has failed"
    ist2[3].mask = np.ones(im.shape, dtype=bool)
    assert np.all(ist2[3].mask), "setting mask on an image stack item not working"
    istack2 = selfistack2.clone
    mask = ImageFile(np.zeros_like(istack2[0].image)).mask.draw.circle(20, 20, 10).mask
    mask = ~mask
    istack2.each.mask = mask
    assert istack2[0].mask[0, 0], "Mask not set correctly in stack"
    assert not istack2[0].mask[20, 20], "Mask not set correctly in stack"


if __name__ == "__main__":
    pytest.main(["--pdb", __file__])



# -*- coding: utf-8 -*-
"""Useful Utility classes."""

__all__ = [
    "attributeStore",
    "typedList",
    "Options",
    "get_option",
    "set_option",
    "itersubclasses",
    "subclasses",
    "copy_into",
]

import copy
from typing import Optional, Dict, Any, List, Iterable as IterableType, Union
from collections.abc import MutableSequence

from .tests import all_type, isiterable

_options = {
    "short_repr": False,
    "short_data_repr": False,
    "short_folder_rrepr": True,
    "short_img_repr": True,
    "no_figs": True,
    "multiprocessing": False,  # Change default to not use multiprocessing for now
    "threading": False,
    "warnings": False,
}


class attributeStore(dict):
    """A dictionary=like class that provides attributes that work like indices.

    Used to implement the mapping of column types to indices in the setas attriobutes.
    """

    def __init__(self, *args: Any, **kargs: Any) -> None:
        """Initialise from a dictionary."""
        if len(args) == 1 and isinstance(args[0], dict):
            self.update(args[0])
        else:
            super().__init__(*args, **kargs)

    def __setattr__(self, name: str, value: Any) -> None:
        """Set an attribute (equivalent to setting an item)."""
        self[name] = value

    def __getattr__(self, name: str) -> Any:
        """Get an attribute (equivalent to getting an item)."""
        try:
            return self[name]
        except KeyError as err:
            raise AttributeError from err


def itersubclasses(cls: type, _seen: Optional[set] = None) -> List[type]:
    """Iterate over subclasses of a given class.

    Generator over all subclasses of a given class, in depth first order.

    >>> list(itersubclasses(int)) == [bool]
    True
    >>> class A(object): pass
    >>> class B(A): pass
    >>> class C(A): pass
    >>> class D(B,C): pass
    >>> class E(D): pass
    >>>
    >>> for cls in itersubclasses(A):
    ...     print(cls.__name__)
    B
    D
    E
    C
    >>> # get ALL (new-style) classes currently defined
    >>> [cls.__name__ for cls in itersubclasses(object)] #doctest: +ELLIPSIS
    ['type', ...'tuple', ...]
    """
    if not isinstance(cls, type):
        raise TypeError(f"itersubclasses must be called with new-style classes, not {cls}")
    if _seen is None:
        _seen = set()
    try:
        subs = cls.__subclasses__()
    except TypeError:  # fails only when cls is type
        subs = cls.__subclasses__(cls)
    for sub in subs:
        if sub not in _seen:
            _seen.add(sub)
            itersubclasses(sub, _seen)
    return list(_seen)


def subclasses(cls: Optional[type] = None) -> Dict:  # pylint: disable=no-self-argument
    """Return a list of all in memory subclasses of this DataFile."""
    # pylint: disable=E1136, E1135
    if cls is None:
        from ..Core import DataFile  # pylint: disable=import-outside-toplevel

        cls = DataFile

    _subclasses = getattr(cls, "_subclasses", None)
    if _subclasses is None:
        _subclasses = dict()
    tmp = itersubclasses(cls)
    if len(_subclasses) < 1 or _subclasses[0] != len(tmp):  # Rebuild index
        tmp = {
            x: (getattr(x, "priority", 256), x.__name__)
            for x in sorted(tmp, key=lambda c: (getattr(c, "priority", 256), getattr(c, "__name__", "None")))
        }
        tmp = {v[1]: k for k, v in tmp.items()}
        ret = dict()
        ret[cls.__name__] = cls
        ret.update(tmp)
        _subclasses = dict()
        _subclasses = (len(tmp), ret)
    else:
        ret = dict(_subclasses[1])
    setattr(cls, "_subclasses", _subclasses)
    return ret


class typedList(MutableSequence):
    """Subclass list to make setitem enforce  strict typing of members of the list."""

    def __init__(self, *args: Any, **kargs: Any) -> None:
        """Construct the typedList."""
        self._store = []
        if (not args) or not (isinstance(args[0], type) or (isinstance(args[0], tuple) and all_type(args[0], type))):
            self._type = str  # Default list type is a string
        else:
            args = list(args)
            self._type = args.pop(0)
        if not args:
            self._store = list(*args, **kargs)
        elif len(args) == 1 and all_type(args[0], self._type):
            self._store = list(*args, **kargs)
        else:
            if len(args) > 1:
                raise SyntaxError("List should be constructed with at most two arguments, a type and an iterable")
            raise TypeError(f"List should be initialised with elements that are all of type {self._type}")

    def __add__(self, other: IterableType) -> "typedList":
        """Add operator works like ordinary lists."""
        if isiterable(other):
            new = copy.deepcopy(self)
            new.extend(other)
            return new
        return NotImplemented

    def __iadd__(self, other: IterableType) -> "typedList":
        """Inplace-add works like a list."""
        if isiterable(other):
            self.extend(other)
            return self
        return NotImplemented

    def __radd__(self, other: IterableType) -> "typedList":
        """Support add on the right like a list."""
        if isinstance(other, list):
            return other + self._store
        return NotImplemented

    def __eq__(self, other: List) -> bool:
        """Equality test."""
        return self._store == other

    def __delitem__(self, index: int) -> None:
        """Remove an item like in a list."""
        del self._store[index]

    def __getitem__(self, index: int) -> Any:
        """Get an item like in a list."""
        return self._store[index]

    def __len__(self) -> int:
        """Implement the len function like a list."""
        return len(self._store)

    def __repr__(self) -> str:
        """Textual representation like a list."""
        return repr(self._store)

    def __setitem__(self, name: Union[int, IterableType, slice], value: Any) -> None:
        """Sett an item and do some type checks."""
        if isiterable(name) or isinstance(name, slice):
            if not isiterable(value) or not all_type(value, self._type):
                raise TypeError(
                    f"Elements of this list should be of type {self._type} and must set "
                    + "the correct number of elements"
                )
        elif not isinstance(value, self._type):
            raise TypeError(f"Elements of this list should be of type {self._type}")
        self._store[name] = value

    def extend(self, values: IterableType) -> None:  # pylint:  disable=arguments-differ
        """Extend the list and do some type checking."""
        if not isiterable(values) or not all_type(values, self._type):
            raise TypeError(f"Elements of this list should be of type {self._type}")
        self._store.extend(values)

    def index(self, value: Any, start: int = 0, end: Optional[int] = None) -> int:  # pylint:  disable=arguments-differ
        """Index works like a list except we support Python 3 optional parameters everywhere."""
        if end is None:
            end = len(self._store)
        return self._store[start:end].index(value) + start

    def insert(self, index: int, value: Any) -> None:  # pylint:  disable=arguments-differ
        """Insert an element and do some type checking."""
        if not isinstance(value, self._type):
            raise TypeError(f"Elements of this list should be of type {self._type}")
        self._store.insert(index, value)


def get_option(name: str) -> bool:
    """Return the option value."""
    if name not in _options.keys():
        raise IndexError(f"{name} is not a valid package option")
    return _options[name]


def set_option(name: str, value: bool) -> None:
    """Set a global package option.

    Args:
        name (str):
            Option Name, one of:
                - short_repr (bool):
                    Instead of using a rich representation, use a short description for DataFile and Imagefile.
                - short_data_repr (bool):
                    Just use short representation for DataFiles
                - short_img_repr (bool):
                    Just use a short representation for image file
                - no_figs (bool):
                    Do not return figures from plotting functions, just plot them.
        value (depends on name):
            The value to set (see *name*)
    """
    if name not in _options.keys():
        raise IndexError(f"{name} is not a valid package option")
    if not isinstance(value, bool):
        raise ValueError(f"{name} takes a boolean value not a {type(value)}")
    _options[name] = value


class Options:
    """Dead simple class to allow access to package options."""

    def __init__(self):
        """Options class wraps the get/set_option calls into an object orientated interface."""
        self._defaults = copy.copy(_options)

    def __setattr__(self, name: str, value: bool) -> None:
        """Set an option value."""
        if name.startswith("_"):
            return super().__setattr__(name, value)
        if name not in _options:
            raise AttributeError(f"{name} is not a recognised option.")
        if not isinstance(value, type(_options[name])):
            raise ValueError(f"{name} takes a {type(_options[name])} not a {type(value)}")
        set_option(name, value)

    def __getattr__(self, name: str) -> bool:
        """Lookup an option value."""
        if name not in _options:
            raise AttributeError(f"{name} is not a recognised option.")
        return get_option(name)

    def __delattr__(self, name: str) -> None:
        """Clear and Option value back to defaults."""
        if name not in _options:
            raise AttributeError(f"{name} is not a recognised option.")
        set_option(name, self._defaults[name])

    def __dir__(self) -> List[str]:
        """Return a list of the aviailable options."""
        return list(_options.keys())

    def __repr__(self) -> str:
        """Make a standard text representation of Options class."""
        s = "Stoner Package Options\n"
        s += "~~~~~~~~~~~~~~~~~~~~~~\n"
        for k in dir(self):
            s += f"{k} : {get_option(k)}\n"
        return s


def copy_into(source: "DataFile", dest: "DataFile") -> "DataFile":
    """Copy the data associated with source to dest.

    Args:
        source(DataFile): The DataFile object to be copied from
        dest (DataFile): The DataFile objrct to be changed by receiving the copiued data.

    Returns:
        The modified *dest* DataFile.

    Unlike copying or deepcopying a DataFile, this function preserves the class of the destination and just
    overwrites the attributes that represent the data in the DataFile.
    """
    dest.data = source.data.copy()
    dest.setas = source.setas
    dest.fig = getattr(source, "fig", None)
    for attr in source._public_attrs:
        if not hasattr(source, attr) or callable(getattr(source, attr)) or attr in ["data", "fig"]:
            continue
        try:
            setattr(dest, attr, copy.deepcopy(getattr(source, attr)))
        except (NotImplementedError, TypeError, ValueError):  # Deepcopying failed, so just copy a reference instead
            try:
                setattr(dest, attr, getattr(source, attr))
            except ValueError:
                pass
    dest._punlic_attrs = source._public_attrs
    return dest



"""Test Weak-localisation fitting."""

# pylint: disable=invalid-name
from copy import deepcopy

from numpy import linspace, ones_like
from numpy.random import normal

from Stoner import Data
from Stoner.analysis.fitting.models.e_transport import (
    blochGrueneisen,
    BlochGrueneisen,
)

T = linspace(4.2, 300, 101)
params = [265, 65, 1.0, 5]
params2 = deepcopy(params)
G = blochGrueneisen(T, *params) + normal(size=len(T), scale=5e-5)
dG = ones_like(T) * 5e-5
d = Data(
    T,
    G,
    dG,
    setas="xye",
    column_headers=["Temperature (K)", "Resistivity", "dR"],
)

d.curve_fit(blochGrueneisen, p0=params, result=True, header="curve_fit")

d.setas = "xy"
d.lmfit(BlochGrueneisen, result=True, header="lmfit")

d.setas = "xyeyy"
d.plot(fmt=["r.", "b-", "g-"])

d.annotate_fit(blochGrueneisen, x=0.65, y=0.35, fontdict={"size": "x-small"})
d.annotate_fit(
    BlochGrueneisen,
    x=0.65,
    y=0.05,
    fontdict={"size": "x-small"},
    prefix="BlochGrueneisen",
)
d.title = "Bloch-Grueneisen Fit"



# -*- coding: utf-8 -*-
"""
Created on Tue Dec 11 09:33:08 2018

@author: phygbu
"""

import sys
import os.path as path
import pytest


pth = path.dirname(__file__)
pth = path.realpath(path.join(pth, "../../../"))
sys.path.insert(0, pth)

from Stoner import Data, __home__, Options
from Stoner.Folders import PlotFolder
from Stoner.plot.formats import TexEngFormatter, DefaultPlotStyle
import matplotlib.pyplot as plt


def extra(i, j, d):
    """Cleanup plot window."""
    d.title = "{},{}".format(i, j)
    d.xlabel("$V$")
    d.ylabel("$I$")


datadir = path.join(pth, "sample-data")


def test_plotting():
    selffldr = PlotFolder(path.join(datadir, "NLIV"), pattern="*.txt", setas="yx")
    Options.multiprocessing = False
    selffldr.plots_per_page = len(selffldr)
    selffldr.plot(figsize=(18, 12), title="{iterator}")
    assert len(plt.get_fignums()) == 1, "Plotting to a single figure in PlotFolder failed."
    selffldr.plot(extra=extra)
    assert len(plt.get_fignums()) == 2, "Plotting to a single figure in PlotFolder failed."
    selfax = selffldr[0].subplots
    assert len(selfax) == 16, "Subplots check failed."

    plt.close("all")
    Options.multiprocessing = False


if __name__ == "__main__":  # Run some tests manually to allow debugging
    pytest.main(["--pdb", __file__])



"""Smoothing Data methods example."""

# pylint: disable=invalid-name, no-member, not-callable
import matplotlib.pyplot as plt

from Stoner import Data

fig = plt.figure(figsize=(9, 6))

d = Data("Noisy_Data.txt", setas="xy")

d.fig = fig
d.plot(color="grey")
# Filter with Savitsky-Golay filter, linear over 7 ppoints
d.SG_Filter(result=True, points=11, header="S-G Filtered")
# Filter with cubic splines
d.spline(result=True, order=3, smoothing=4, header="Spline")
# Rebin data
d.smooth("hamming", size=0.2, result=True, header="Smoothed")

d.plot(0, "S-G Filtered", lw=2, label="SG Filter")
d.plot(0, "Spline", lw=2, label="Spline")
d.plot(0, "Smooth", lw=2, label="Smoooth", color="green")

d2 = d.bin(bins=100, mode="lin")
d2.fig = d.fig
d2.plot(lw=2, label="Re-binned", color="blue")
d2.xlim(3.5, 6.5)
d2.ylim(-200.0, 400.0)



"""USe curve_fit to fit a straight line."""

# pylint: disable=invalid-name
from Stoner import Data


def linear(x, m, c):
    """Straight line function."""
    return m * x + c


d = Data("curve_fit_data.dat", setas="xye")
d.plot(fmt="ro")
fit = d.curve_fit(
    linear, result=True, replace=False, header="Fit", output="report"
)
d.setas = "x..y"
d.plot(fmt="b-")
d.annotate_fit(linear)
d.draw()
print(fit.fit_report())



# -*- coding: utf-8 -*-
"""Teest stoner.core.operators module"""

import pytest
from os import path
import re
import numpy as np

from Stoner import Data, __home__

pth = path.dirname(__file__)
pth = path.realpath(path.join(pth, "../../"))
datadir = path.join(pth, "sample-data")

selfd = Data(path.join(path.dirname(__file__), "..", "CoreTest.dat"), setas="xy")


def test_floordiv_operators():
    selfd = Data(path.join(path.dirname(__file__), "..", "CoreTest.dat"), setas="xy")

    assert all(selfd // 0 == selfd.column(0)), "Failed the // operator with integer"
    assert all(selfd // "X-Data" == selfd.column(0)), "Failed the // operator with string"
    assert all((selfd // re.compile(r"^X\-"))[:, 0] == selfd.column(0)), "Failed the // operator with regexp"
    with pytest.raises(TypeError):
        selfd + 2.2


def test_eq_operator():
    selfd = Data(path.join(path.dirname(__file__), "..", "CoreTest.dat"), setas="xy")
    selfd2 = selfd.clone
    selfd3 = selfd.clone
    selfd3.data = selfd3.data[:, 1]
    assert selfd == selfd2
    assert not selfd == 1
    assert not selfd3 == selfd
    selfd3 = selfd.clone
    selfd3.column_headers[0] = "Different"
    assert not selfd == selfd3
    selfd3 = selfd.clone
    selfd3["Extra"] = False
    assert not selfd == selfd3


def test_modulo_operators():
    selfd = Data(path.join(path.dirname(__file__), "..", "CoreTest.dat"), setas="xy")
    t = [selfd % 1, selfd % "Y-Data", (selfd % re.compile(r"Y\-"))]
    for ix, tst in enumerate(["integer", "string", "regexp"]):
        assert all(t[ix].data == np.atleast_2d(selfd.column(0)).T), f"Failed % operator with {tst} index"
    d = selfd % 1
    assert d.shape[1] == 1, "Column division failed"
    assert len(d.setas) == d.shape[1], "Column removal messed up setas"
    assert len(d.column_headers) == d.shape[1], "Column removal messed up column headers"


def test_and_operator():
    selfd = Data(path.join(path.dirname(__file__), "..", "CoreTest.dat"), setas="xy")
    selfd2 = Data(path.join(__home__, "..", "sample-data", "TDI_Format_RT.txt"))
    d = selfd & selfd.x
    assert d.shape[1] == 3, "& operator failed."
    assert len(d.setas) == 3, "& messed up setas"
    assert len(d.column_headers) == 3, "& messed up setas"
    d &= d.x
    assert d.shape[1] == 4, "Inplace & operator failed"
    e = selfd.clone
    f = selfd2.clone
    g = e & f
    h = f & e
    assert g.shape == h.shape, "Anding unequal column lengths faile!"


def test_add_operator():
    selfd = Data(path.join(path.dirname(__file__), "..", "CoreTest.dat"), setas="xy")
    selfd2 = Data(path.join(__home__, "..", "sample-data", "TDI_Format_RT.txt"))
    empty = Data()
    empty = empty + np.zeros(10)
    assert empty.shape == (1, 10), "Adding to an empty array failed"
    d = selfd + np.array([0, 0])
    assert len(d) == len(selfd) + 1, "+ operator failed."
    d += np.array([0, 0])
    assert len(d) == len(selfd) + 2, "Inplace + operator failed."
    e = selfd.clone
    f = selfd2.clone
    g = e + f
    assert g.shape == (1776, 2), "Add of 2 column and 3 column failed"
    g = f + e
    assert g.shape == (1776, 3), "Add of 3 column and 2 column failed."
    f = selfd2.clone
    f += {"Res": 0.3, "Temp": 1.2, "New_Column": 25}
    f1, f2 = f.shape
    assert f1 == selfd2.shape[0] + 1, "Failed to add a row by providing a dictionary"
    assert f2 == selfd2.shape[1] + 1, "Failed to add an extra columns by adding a dictionary with a new key"
    assert np.isnan(f[-1, 2]) and np.isnan(f[0, -1]), "Unset values when adding a dictionary not NaN"
    d = Data()
    d += np.ones(5)
    d += np.zeros(5)
    assert np.all(d // 1 == np.array([1, 0])), "Adding 1D arrays should add new rows"
    with pytest.raises(TypeError):
        d += True


def test_delete_operator():
    selfd = Data(path.join(path.dirname(__file__), "..", "CoreTest.dat"), setas="xy")
    selfd2 = Data(path.join(__home__, "..", "sample-data", "TDI_Format_RT.txt"))
    d = selfd + np.array([0, 0])
    d += np.array([0, 0])
    d = d - (-1)
    assert len(d) == len(selfd) + 1, f"Delete with integer index failed. {len(d)} vs {len(selfd)+1}"
    d -= -1
    assert len(d) == len(selfd), f"Inplace delete with integer index failed. {len(d)} vs {len(selfd)}"
    d -= slice(0, -1, 2)
    assert len(d) == len(selfd) / 2, f"Inplace delete with slice index failed. {len(d)} vs {len(selfd)/2}"

    e = selfd.clone
    f = selfd2.clone
    g = e + f
    assert g.shape == (1776, 2), "Add of 2 column and 3 column failed"
    g = f + e
    assert g.shape == (1776, 3), "Add of 3 column and 2 column failed."
    g = e & f
    h = f & e
    assert g.shape == h.shape, "Anding unequal column lengths faile!"


def test_invert_operator():
    selfd = Data(path.join(path.dirname(__file__), "..", "CoreTest.dat"), setas="xy")
    e = ~selfd
    assert e.setas[0] == "y", "failed to invert setas columns"
    d = Data(np.array([[1, 2, 3, 4]]), setas="xdyz")
    assert (~d).setas == ["y", "e", "z", "x"], "Inverting multi-axis array failed"
    with pytest.raises(TypeError):
        Data() << 45.3


def test_lshoft_operator():
    with open(selfd.filename, "r") as data:
        lines = data.readlines()
    e = Data() << lines
    e.setas = selfd.setas
    e["Loaded as"] = selfd["Loaded as"]
    assert e == selfd, "Failed to read and create from << operator"
    with open(selfd.filename, "r") as data:
        e = Data() << data
    e.setas = selfd.setas
    e["Loaded as"] = selfd["Loaded as"]
    assert e == selfd, "Failed to read and create from << operator"


if __name__ == "__main__":
    pytest.main(["--pdb", __file__])



# -*- coding: utf-8 -*-
"""Test the HDF5 file format handling."""


import os.path as path
import tempfile
import Stoner
import pytest
import Stoner.HDF5 as SH

Data = Stoner.Data

pth = path.dirname(__file__)
testdata = path.realpath(path.join(pth, "test-data"))

root = path.realpath(path.join(Stoner.__home__, ".."))
sample_data = path.realpath(path.join(root, "sample-data", "NLIV"))
tmpdir = tempfile.mkdtemp()


def test_HDF5folder():
    # Test constructor from DataFolder
    self_fldr = Stoner.DataFolder(sample_data, pattern="*.txt")
    self_HDF5fldr = SH.HDF5Folder(self_fldr)
    assert self_fldr.shape == self_HDF5fldr.shape, "HDF5Folder created from DataFolder didn't keep the same shape"
    assert self_fldr[0] == self_HDF5fldr[0], "First element of HDF5Folder created from DataFolder changed!"
    HDF5name = path.join(tmpdir, "test-HDF5folder.HDF5")
    self_HDF5fldr.save(HDF5name)
    assert self_fldr.shape == self_HDF5fldr.shape, "HDF5Folder Changed shape when saving!"
    self_HDF5fldr_2 = SH.HDF5Folder(HDF5name)
    self_HDF5fldr_2.compress()
    self_HDF5fldr.sort("i")
    self_HDF5fldr_2.sort("i")
    assert (
        self_HDF5fldr_2.shape == self_HDF5fldr.shape
    ), "HDF5Folder loaded from disc not same shape as HDF5Folder in memory!"
    self_h1 = self_HDF5fldr[0]
    self_h2 = self_HDF5fldr_2[0]
    self_h2.metadata["Stoner.class"] = "Data"  # Correct the loader class
    self_h2.metadata["Loaded from"] = self_h1.metadata["Loaded from"]  # Corrects a path separator bug on Windows
    assert self_h1 == self_h2, "File from loaded HDF5Folder not the same as in memory HDF5Folder."


if __name__ == "__main__":  # Run some tests manually to allow debugging
    pytest.main(["--pdb", __file__])



"""Test Weak-localisation fitting."""

# pylint: disable=invalid-name
from numpy import logspace, ones_like, log10
from numpy.random import normal

from Stoner import Data
from Stoner.analysis.fitting.models.e_transport import (
    fluchsSondheimer,
    FluchsSondheimer,
)

B = logspace(log10(2), 2, 26)
params = [12.5, 0.75, 1e3]
G = fluchsSondheimer(B, *params) + normal(size=len(B), scale=5e-5)
dG = ones_like(B) * 5e-5
d = Data(
    B,
    G,
    dG,
    setas="xye",
    column_headers=["Thickness (nm)", "Conductance", "dConductance"],
)

d.curve_fit(fluchsSondheimer, p0=params, result=True, header="curve_fit")

d.setas = "xye"
d.lmfit(FluchsSondheimer, result=True, header="lmfit")

d.setas = "xyeyy"
d.plot(fmt=["r.", "b-", "g-"])

d.annotate_fit(
    fluchsSondheimer,
    x=0.2,
    y=0.6,
    fontdict={"size": "x-small", "color": "blue"},
)
d.annotate_fit(
    FluchsSondheimer,
    x=0.2,
    y=0.4,
    fontdict={"size": "x-small", "color": "green"},
    prefix="FluchsSondheimer",
)
d.title = "Fluchs-Sondheimer Fit"



"""Example plot using experimental GBStyle."""

# pylint: disable=invalid-name
import os.path as path

from Stoner import Data, __home__
from Stoner.plot.formats import GBPlotStyle

filename = path.realpath(
    path.join(__home__, "..", "doc", "samples", "sample.txt")
)
d = Data(filename, setas="xy", template=GBPlotStyle)
d.y = d.y - (max(d.y) / 2)
d.plot()



"""Example of using lmfit to do a bounded fit."""

# pylint: disable=invalid-name
from Stoner import Data
from Stoner.analysis.fitting.models.generic import StretchedExp

# Load dat and plot
d = Data("lmfit_data.txt", setas="xy")
d.y = d.y

# Do the fit
d.lmfit(StretchedExp, result=True, header="Fit", residuals=True)
# plot
d.setas = "xyy"

d.plot(fmt=["+", "-"])
# Make apretty label using Stoner.Util methods
text = r"$y=A e^{-\left(\frac{x}{x_0}\right)^\beta}$" + "\n"
text += d.annotate_fit(StretchedExp, text_only=True)
d.text(4, 6e4, text, fontsize="x-small")
